%%
%% benchmark tools
%%

functor

import
   Property System(gcDo)

export
   getTime: GetTime
   simple:  SimpleBenchmark
   stat:    Stat
   loop1:   LoopBenchmark
   loop2:   Benchmark
   pprint:  PPrint
   
define
   
   fun {GetTime}
      {Record.filter {Property.get 'time'} IsInt}
   end
   
   %%
   %% repeat {Proc} N times
   %%
   fun {SimpleBenchmark N Proc}
      T1={GetTime} {For 1 N 1 Proc} T2={GetTime} in
      {Record.zip T2 T1 Number.'-'}
   end

   %%
   %% try to get more precise results by calculating average time
   %%
   class Stat
      attr n min max sum
      meth init(DummyVal infinity:Inf<={Pow 2 1000})
         n   <- 0
         min <- {Record.map DummyVal fun {$ _} Inf end}
         max <- {Record.map DummyVal fun {$ _} 0 end}
         sum <- {Record.map DummyVal fun {$ _} 0 end}
      end
      meth val(X)
         n   <- @n+1
         min <- {Record.zip @min X Min}
         max <- {Record.zip @max X Max}
         sum <- {Record.zip @sum X Number.'+'}
      end
      meth get($)
         Avg  = {Record.map @sum fun {$ N} N div @n end} in
         stat(n:@n min:@min max:@max sum:@sum avg:Avg) 
      end
   end

   %%
   %% inner loop
   %%
   fun {LoopBenchmark K Proc N}
      St = {New Stat init({GetTime})} in
      {For 1 N 1 proc {$ _}
                    {St val({SimpleBenchmark K Proc})}
                 end}
      {St get($)}
   end
   
   %%
   %% outer loop
   %%
   proc {Skip1 _} skip end
   %%
   fun {Benchmark K Proc N M}
      {For 1 N*M*K 1 Skip1}
      {System.gcDo}
      St = {New Stat init({GetTime})}
      {For 1 M 1 proc {$ _}
                    {St val({LoopBenchmark K Proc N}.sum)}
                 end}
      Result = {Record.subtract {St get($)} n}
   in
      {Adjoin unit(n:N m:M k:K)
       {Record.map Result
        fun {$ R} {Record.map R fun {$ I} I div N end} end}}
   end

   %%
   %% pretty print function for results
   %%
   local
      fun {Spaces N}
         {ForThread 1 N 1 fun {$ Tail _} & |Tail end nil}
      end
      LenVS = VirtualString.length
      fun {MaxLength VSs}
         {FoldL {Map VSs LenVS} Max 0}
      end
      fun {Right VS N}
         {Spaces N-{LenVS VS}}#VS
      end
      /*
      fun {Left VS N}
         VS#{Spaces N-{LenVS VS}}
      end
      */
      fun {Concat L}
         case L of nil then ''
         [] VS|R then VS#{Concat R} end
      end
   in          
      fun {PPrint X Features}
         Ar   = {Arity X.max}
         MaxL = {MaxLength Ar}
         L    = 8
         Feat = case Features of all then Ar
                else Features end
      in
         '\n'
         #X.m#" x "#X.n#" tests with "#X.k#" procedure calls each"#'\n'
         #'\n'
         #{Spaces MaxL+1}#' '
         #' '#{Right 'min' L}
         #' '#{Right 'avg' L}
         #' '#{Right 'max' L}
         %%#' \t'#{Right 'sum' L}
         #'\n'
         #{Concat {Map Feat
                   fun {$ F}
                      F#{Spaces MaxL-{LenVS F}+1}#':'
                      #' '#{Right X.min.F L}
                      #' '#{Right X.avg.F L}
                      #' '#{Right X.max.F L}
                      %%#' \t'#{Right X.sum.F L}
                      #'\n'
                   end}}
         #'\n'
      end
   end
end

