%%
%% buffer classes
%%

functor

import
   System(eq:Eq /*show:Show*/ showInfo:ShowInfo)
   
export
   simple: Simple
   lifo: LIFO
   fifo: FIFO
   collector: Collector
   
define

   class Simple
      prop
         locking
      attr
         content: nil
         
      meth init
         content <- nil
      end
      
      meth put(S)
         lock
            content <- {Append @content S}
         end
      end
      
      meth getS($)
         lock
            if @content == nil then
               false
            else
               Line Rest
            in
               {String.token @content &\n Line Rest}
               content <- Rest
               Line
            end
         end
      end
   end %class
   
   class LIFO
      prop
         locking
      attr
         stack:nil
         
      meth init
         skip
      end
      meth put(X)
         lock
            stack <- X|@stack
         end
      end
      meth get(?X)
         lock
            if @stack==nil
            then X=false
            else
               !X|R = @stack in
               stack <- R
            end
         end
      end
      meth getAll(?L)
         lock
            L = @stack
            stack <- nil
         end
      end
   end

   class FIFO
      prop
         locking
      attr
         list
         tail
         
      meth init
         L in
         list <- L
         tail <- L
      end
      meth put(X)
         lock
            T in
            @tail = X|T
            tail <- T
         end
      end
      meth get(?X)
         lock
            if {Eq @list @tail}
            then X=false
            else
               !X|R = @list in
               list <- R
            end
         end
      end
      meth getAll(?L)
         lock
            @tail = nil
            L = @list
            FIFO, init
         end
      end
   end
   
   class Collector
      prop
         locking
      feat
         Wait
         ReturnProc
      attr
         NWait
         Idle

      meth init(Return) = M
         %%{Show M}
         self.Wait       = {NewDictionary}
	 self.ReturnProc = case {Value.type Return} of procedure then Return
                           [] port   then proc {$ X} {Port.send Return X} end
                           [] object then proc {$ X} {Return put(X)} end
                           end
         NWait <- 0
         Idle  <- unit
      end
      meth expect(ID R) = M
         %%{Show M}
         lock
            if {Arity R} == nil then
               {self.ReturnProc ID#R}
            else
               {Dictionary.put self.Wait ID
                {NewCell {Record.width R}}#{Record.map R fun {$ X} X#_ end}}
               NWait <- @NWait + 1
               if {IsDet @Idle} then Idle<-_ end
            end
         end
      end
      meth add(ID Feat X) = M
         %%{Show M}
         lock
            try
               C#R = {Dictionary.get self.Wait ID}
               R.Feat.2 = X
            in
               if {Label R}==oneOf then
                  R1 = {MakeRecord oneOf [Feat]} in
                  R1.Feat = R.Feat
                  {self Return(ID R1)}
               else
		  N = {Access C}-1 in
                  {Assign C N}
                  if N==0 then {self Return(ID R)} end
               end
            catch system(kernel(dict...)...) then
               {ShowInfo "Collector: Key `"#ID#"' not found... ignored."}
            end
         end
      end
      
      meth Return(ID R) = M
         {self.ReturnProc ID#R}
         {Dictionary.remove self.Wait ID}
         NWait <- @NWait-1
         if @NWait==0 then @Idle=unit end
      end
      meth idle($) = M
         %%{Show M}
         lock
            {IsDet @Idle}
         end
      end
      meth current($) = M
         %%{Show M}
         lock
	    {Dictionary.entries self.Wait}
         end
      end
      meth terminate(?Current<=_) = M
         %%{Show M}
         lock
            Collector, current(?Current)
            {Thread.terminate {Thread.this}}
         end
      end
   end
end





