%%%
%%% Author:
%%%   Andreas Franke <afranke@ags.uni-sb.de>
%%%
%%% Copyright:
%%%   Andreas Franke, 2000
%%%
%%% Last change:
%%%   $Date: 2001/07/13 13:47:20 $ by $Author: afranke $
%%%   $Revision: 1.1 $
%%%
%%% This file is part of MathWeb
%%%    http://www.mathweb.org/mathweb
%%%
%%% See the file "LICENSE" or
%%%    http://www.mathweb.org/mathweb/LICENSE.html
%%% for information on usage and redistribution
%%% of this file, and for a DISCLAIMER OF ALL
%%% WARRANTIES.
%%%

%%%
%%% CRC
%%%

functor

import
   Open(file)
   OS(tmpnam unlink)
   Pickle(pack save)
   
export
   compute:     CRC

   fromValue1:  CRC1         % using Pickle.save
   fromValue2:  CRC2         % using Pickle.pack and CRC
   
   picklePack1: PicklePack1  % using Pickle.save
   picklePack2: PicklePack2  % using Pickle.pack and CRC
   
define
   
   proc {LoadS File ?S}
      F = {New Open.file init(name:File)}
   in
      {F read(list:?S size:all)}
      {F close}
   end
   proc {PicklePack1 X ?BS}
      Tmp = {OS.tmpnam} in
      {Pickle.save X Tmp}
      {ByteString.make {LoadS Tmp} ?BS}
      {OS.unlink Tmp}
   end
   fun {CRC1 X}
      {List.take {List.drop {VirtualString.toString {PicklePack1 X}} 3} 4}
   end
   
   Bits = 32
   Bits1= Bits - 1
   %%
   fun {NewVal}
      {BitArray.new 0 Bits1}
   end
   Values = {Array.new 0 Bits1 0}
   local
      proc {Loop I N}
         if I<Bits then
            {Array.put Values I N}
            {Loop I+1 2*N}
         end
      end
   in
      {Loop 0 1}
   end
   NULL = {NewVal}
   FULL = {NewVal}
   {For 0 Bits1 1
    proc {$ I} {BitArray.set FULL I} end}
   %%
   proc {NOT X ?Y}
      Y = {BitArray.clone FULL}
      {BitArray.nimpl Y X}
   end
   proc {OR X Y ?Z}
      Z = {BitArray.clone X}
      {BitArray.disj Z Y}
   end
   proc {AND X Y ?Z}
      Z = {BitArray.clone X}
      {BitArray.conj Z Y}
   end
   fun {XOR X Y}
      {AND {OR X Y} {NOT {AND X Y}}}
   end
   %%
   proc {RShift X N ?Y}
      Y = {NewVal}
      {For N Bits1 1
       proc {$ I}
          if {BitArray.test X I} then
             {BitArray.set Y I-N}
          end
       end}
   end
   
   %%
   %% conversion BitArray <--> Int
   %%
   local
      proc {Loop I N X}
         if I<Bits then 
            if (N mod 2)==1 then
               {BitArray.set X I}
            end
            {Loop I+1 (N div 2) X}
         end
      end
   in
      proc {FromInt N ?X}
         X = {NewVal}
         {Loop 0 N X}
      end
   end
   fun {ToInt X}
      {FoldL {Map {BitArray.toList X} fun {$ I} {Array.get Values I} end} 
       Number.'+' 0}
   end

   /*
   fun {Not A}   {ToInt {NOT {FromInt A}}} end
   fun {And A B} {ToInt {AND {FromInt A} {FromInt B}}} end
   fun {Or A B}  {ToInt {OR  {FromInt A} {FromInt B}}} end
   fun {Xor A B} {ToInt {XOR {FromInt A} {FromInt B}}} end
   */
   /*
   {And 5 9}
   {Or 5 9}
   {Xor 5 9}
   {Not 5}
   */
   
   %%
   %% hex --> dec 
   %%
   local
      HexDigits = digits(&0:0 &1:1 &2:2 &3:3 &4:4 &5:5 &6:6 &7:7 &8:8 &9:9
                         &a:10 &b:11 &c:12 &d:13 &e:14 &f:15
                         &A:10 &B:11 &C:12 &D:13 &E:14 &F:15)
      fun {Loop Is Acc}
         case Is of nil then Acc
         [] I|Ir then {Loop Ir Acc*16+HexDigits.I}
         end
      end
   in
      fun {HexToInt S}
         {Loop S 0}
      end
   end

%%% The following C-code has been taken from 
%%%         mozart/platform/emulator/pickle.cc
%%%
%%% The original source can be found e.g. here:
%%%         http://www.w3c.rl.ac.uk/isopng/D-PNG-CRCAppendix.html

%%% // The following sample code represents a practical implementation of the
%%% // CRC (Cyclic Redundancy Check) employed in PNG chunks. (See also ISO
%%% // 3309 [ISO-3309] or ITU-T V.42 [ITU-V42] for a formal specification.) 
%%%
%%% /* Table of CRCs of all 8-bit messages. */
%%% crc_t crc_table[256];

   CRC_Table = {Array.new 0 255 unit}
   
%%% /* Make the table for a fast CRC. */
%%% void make_crc_table(void) 
%%% {
%%%   crc_t c;
%%%   int n, k;
%%%    
%%%   for (n = 0; n < 256; n++) {
%%%     c = (crc_t) n;
%%%     for (k = 0; k < 8; k++) {
%%%       if (c & 1)
%%%         c = 0xedb88320L ^ (c >> 1);
%%%       else
%%%         c = c >> 1;
%%%     }
%%%     crc_table[n] = c;
%%%   }
%%% 
%%% }
   
   EDB88320 = {FromInt {HexToInt "edb88320"}}
 
   proc {Make_CRC_Table}
      C = {NewCell {NewVal}} in
      {For 0 255 1
       proc {$ N}
          {Assign C {FromInt N}}
          {For 0 7 1
           proc {$ _}
              if {BitArray.test {Access C} 0}
              then {Assign C {XOR EDB88320 {RShift {Access C} 1}}}
              else {Assign C               {RShift {Access C} 1}}
              end
           end}
          {Array.put CRC_Table N {BitArray.clone {Access C}}}
       end}
   end
   {Make_CRC_Table}
   
%%% /* Update a running CRC with the bytes buf[0..len-1]--the CRC
%%%   should be initialized to all 1's, and the transmitted value
%%%   is the 1's complement of the final running CRC (see the
%%%   crc() routine below)). */
%%%
%%% crc_t update_crc(crc_t crc, unsigned char *buf, int len) 
%%% {
%%%  static int tablemade = 0;
%%%  if (!tablemade) {
%%%    make_crc_table();
%%%    tablemade = 1;
%%%  }
%%%
%%%   crc_t c = crc;
%%%   int n;
%%% 
%%%   for (n = 0; n < len; n++) {
%%%     c = crc_table[(c ^ buf[n]) & 0xff] ^ (c >> 8);
%%%   }
%%%   return c;
%%% }

   %%FFFFFFFF = {FromInt {HexToInt "ffffffff"}}
   FF       = {FromInt {HexToInt "ff"}}
   
   fun {Update_CRC CRC BS}
      C = {NewCell CRC} in
      {For 0 {ByteString.length BS}-1 1
       proc {$ N}
          Buf_N = {FromInt {ByteString.get BS N}}
          Index = {ToInt {AND {XOR {Access C} Buf_N} FF}}
       in
          {Assign C {XOR {Array.get CRC_Table Index} {RShift {Access C} 8}}}
       end}
      {Access C}
   end
   
%%% /* Return the CRC of the bytes buf[0..len-1]. */
%%% unsigned long crc(unsigned char *buf, int len)
%%% {
%%%   return update_crc(0xffffffffL, buf, len) ^ 0xffffffffL;
%%% }
   
   fun {CRC BS}
      C0 = {ToInt {Update_CRC NULL BS}}
      %%C0 = {ToInt {XOR {Update_CRC FFFFFFFF BS} FFFFFFFF}}
      B0 = C0 mod 256
      C1 = C0 div 256
      B1 = C1 mod 256
      C2 = C1 div 256
      B2 = C2 mod 256
      C3 = C2 div 256
      B3 = C3 mod 256
   in
      [B0 B1 B2 B3]
   end
   
   fun {CRC2 X}
      {CRC {Pickle.pack X}}
   end
   fun {PicklePack2 X}
      BS = {Pickle.pack X} in
      "\002\002\002"#{CRC BS}#BS
   end

   /*
   fun {FromList [B0 B1 B2 B3]}
      {FromInt ((B0*16+B1)*16+B2)*16+B3}
   end
   */
   /*
   {BitArray.toList {Update_CRC FFFFFFFF /*{FromList {CRC1 unit}}*/ {Pickle.pack unit}}}
   */
end




