%%
%% DaVinci interface (graph visualization)
%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%% DaVinci Syntax %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%% -------------------------------
%%% Term Representation for Graphs:
%%% -------------------------------

% graphterm     ::= [graphterm1] | [].      % Graph with root nodes.
% graphterm1    ::= node | node,graphterm1.
% node          ::=
%     l(node_id,n(type,attributes,edges))   % Node with ID/type/attr/children.
%   | r(node_id).                           % Reference to a node.
%
% edges         ::= [edges1] | [].          % Edges with ID/type/attr/child.
% edges1        ::= edge | edge,edges1.
% edge          ::= 
%     l(edge_id,e(type,attributes,node)).
%
% attributes    ::= [attributes1] | [].     % Attributes (key/value).
% attributes1   ::= attribute | attribute,attributes1.
% attribute     ::= 
%     a(string,string)                 % Normal key/value attribute pair.
%   | m(menus).                        % Popup menu attribute for a node/edge.

%%% --------------
%%% Graph Updates:
%%% --------------

% mixed_updates  ::= [mixed_updates1] | [].  % Delete or remove nodes & edges.
% mixed_updates1 ::= mixed_update | mixed_update,mixed_updates1.
% mixed_update   ::= node_update | edge_update.
% 
% node_updates  ::= [node_updates1] | [].         % Delete or remove nodes.
% node_updates1 ::= node_update | node_update,node_updates1.
% node_update   ::=
%     delete_node(node_id)
%   | new_node(node_id,type,attributes).
% 
% edge_updates  ::= [edge_updates1] | [].         % Delete or remove edges.
% edge_updates1 ::= edge_update | edge_update,edge_updates1.
% edge_update   ::= 
%     delete_edge(edge_id)
%   | new_edge(edge_id,type,attributes,node_id,node_id).
%   | new_edge_behind(edge_id,edge_id,type,attributes,node_id,node_id).
% 
% attr_changes  ::= [attr_changes1] | [].         % Change attributes.
% attr_changes1 ::= attr_change | attr_change,attr_changes1.
% attr_change   ::= 
%     node(node_id,attributes)
%   | edge(edge_id,attributes).
% 
% type_changes  ::= [type_changes1] | [].         % Change types.
% type_changes1 ::= type_change | type_change,type_changes1.
% type_change   ::= 
%     node(node_id,type)
%   | edge(edge_id,type).

%%% ----------------------------
%%% Application Menus and Icons:
%%% ----------------------------

% menus      ::= [menus1] | [].                   % Create Menus.
% menus1     ::= menu_entry | menu_entry,menus1.
% menu_entry ::= 
%     menu_entry(menu_id,menu_label)
%   | menu_entry_mne(menu_id,menu_label,menu_mne,menu_mod,menu_acc)
%   | menu_entry_disabled(menu_id,menu_label)
%   | submenu_entry(menu_id,menu_label,menus)
%   | submenu_entry_mne(menu_id,menu_label,menus,menu_mne)
%   | submenu_entry_disabled(menu_id,menu_label,menus)
%   | blank.
% 
% icons      ::= [icons1] | [].                   % Create Icons.
% icons1     ::= icon_entry | icon_entry,icons1.
% icon_entry ::= 
%     icon_entry(icon_id,filename,string)
%   | blank.

%%% --------------------------
%%% Visualization definitions:
%%% --------------------------

% visual_rules  ::= [visual_rules1] | [].         % Visualization definitions.
% visual_rules1 ::= visual_rule | visual_rule,visual_rules1.
% visual_rule   ::= 
%     nr(type,attributes)
%   | er(type,attributes).

%%% -------------
%%% String Sorts:
%%% -------------

% node_ids   ::= [node_ids1] | [].                % Many node ID's.
% node_ids1  ::= node_id | node_id,node_ids1.
% menu_ids   ::= [menu_ids1] | [].                % Many menu ID's.
% menu_ids1  ::= menu_id | menu_id,menu_ids1.
% icon_ids   ::= [icon_ids1] | [].                % Many icon ID's.
% icon_ids1  ::= icon_id | icon_id,icon_ids1.
%
% node_id    ::= string.  % Unique node ID.
% edge_id    ::= string.  % Unique edge ID.
% menu_id    ::= string.  % Unique menu ID.        
% menu_label ::= string.  % Text of menu entry.    
% menu_mne   ::= string.  % Motif mnemonic char.   
% menu_acc   ::= string.  % Motif accelerator key. 
% icon_id    ::= string.  % Unique icon ID.        
% type       ::= string.  % Arbitrary type.        
% filename   ::= string.  % Valid filename.        
% context_id ::= string.  % Context ID.            
% window_id  ::= string.  % Window ID.             
% string     ::= ... /*Informally: Character sequence without newline, 
%                      carriage return and double quotes, enclosed by double 
%                      quotes ("). For example: "This is a string". 
%                      Special characters can be written in C-notation, e.g. 
%                      \n for newline or \" for a quote inside the string.*/

%%% ------------
%%% Basic Sorts:
%%% ------------

% integer    ::= ... /*Informally: Non-floating positive number.*/
% boolean    ::= true | false.
% orientation::= top_down | bottom_up | left_right | right_left.
% direction  ::= up | down | left | right.
% btypes     ::= [btypes1].               /*Types for file browser. */
% btypes1    ::= btype | btype,btypes1.
% btype      ::= bt(string,string,string)./*Text, pattern and title postfix.*/
% menu_mod   ::= alt | shift | control | meta | none.
%                                         /*Motif modifier key. */

%% usage: _ = {New DaVinciGraph init(G)}
%%
%% where  G = graph(nodes:[a b c d]
%%                  edges:[a#d b#d c#d])
%%
%% or     G = children(a:[d] b:[d] c:[d] d:nil)
%%

functor

prepare
   
   local
      proc {DictAdd D Key X}
         L = {Dictionary.get D Key} in
         {Dictionary.put D Key X|L}
      end
   in
      proc {MakeChildrenDict Nodes Edges ?Children}
         Children = {NewDictionary}
         {ForAll Nodes proc {$ N} {Dictionary.put Children N nil} end}
         {ForAll Edges proc {$ From#To} {DictAdd Children From To} end}
      end
      fun {MakeChildrenRec Nodes Edges}
         {Dictionary.toRecord children {MakeChildrenDict Nodes Edges}}
      end
   end
   local
      fun {MakeEdge Target}
         e('' nil r(Target))
      end
   in
      fun {MakeNode ID Text Children}
         l(ID n('' [a('OBJECT' Text)] {Map Children MakeEdge}))   
      end
   end
   fun {MakeGraph ChildrenRec}
      {Map {Record.toListInd ChildrenRec}
       fun {$ Node#Children}
          {MakeNode Node Node Children}
       end}
   end

import
   DaVinci(daVinciClass: DaVinciClass) at 'x-oz://contrib/DaVinci.ozf'
   
export
   'class': DaVinciGraph
   
define
   
   class DaVinciManager from DaVinciClass
                           
      /*
      attr
         DaVinci: false
         lastdata: 'none'  
         lastnode: 'none'
         lastedge: 'none'
         
      /** create a pipe to the API of DaVinci. */
      /** start a new DaVinci window.          */
      meth open
         case @DaVinci of false then
            DAVINCIHOME = {OS.getEnv 'DAVINCIHOME'} in
            DaVinci <- {New Open.pipe init(cmd:  DAVINCIHOME#'/daVinci'
                                           args: ['-pipe'])}
            %%{Browse DAVINCIHOME#'/daVinci'}
         else skip end
      end
      
      /** close a DaVinci object created by init */
      meth close
         case @DaVinci of false then skip
         else
            DaVinciManager, command(menu "file(close)")
            {@DaVinci close}
            DaVinci <- false
         end
      end
      */
      
      /** send a command to DaVinci                             */
      /** Input: Kind  = what kind of command (menu,graph etc.) */
      /**         Cmd  = the command itself as a string         */
      meth sendCommand(Kind Cmd)
         DaVinciClass, sendVS(Kind#'('#Cmd#')')
      end
      
      /** add a node "Name" to the current graph */
      meth addNode(Name)
         Cmd = graph(update([new_node(Name "Node" [a("OBJECT" Name)])] nil)) in
         DaVinciManager, send(Cmd)
      end
      
      /** add an edge "Name" from "Node1" to "Node2" to the current graph */
      meth addEdge(Name Node1 Node2)
         Cmd = graph(update(nil [new_edge(Name "Edge" nil Node1 Node2)])) in 
         DaVinciManager, send(Cmd)
      end
      
      /*
      %%
      meth node_selected(?Node)
         Node=@lastnode
      end
      
      /** read the output of the DaVinci API in a loop */
      meth readloop
         lastdata <- {ReadDV @DaVinci}
         case @lastdata of "quit\n" then
            lastnode <- 'none'
            DaVinci  <- false
            %%{Browse "DaVinci wurde beendet."}
         [] "node_double_click\n" then
            %%{Browse "Doppelklick auf Knoten "#@LastNode#"!"}
            skip
         [] "edge_double_click\n" then
            %%{Browse "Doppelklick auf Kante "#@LastEdge#"!"}
            DaVinciManager,readloop
         [] &n|&o|&d|&e|&_|&s|&e|&l|&e|&c|&t|&i|&o|&n|&s|&_|&l|&a|&b|&e|&l|&s|&(|R then
            case R of "[])\n" then
               %%{Browse "Klick auf leeres Feld !"}
               DaVinciManager,readloop
            [] &[|&"|R2 then
               lastnode <- {Reverse {List.drop {Reverse R2} 4}}
               %%{Browse  "Der Knoten "#@LastNode#" wurde ausgewhlt!"}
               DaVinciManager,readloop
            else %%{Browse "Fehler!"}
               DaVinciManager,readloop
            end
         [] &e|&d|&g|&e|&_|&s|&e|&l|&e|&c|&t|&i|&o|&n|&_|&l|&a|&b|&e|&l|&(|&"|R then
            lastedge <- {Reverse {List.drop {Reverse R} 3}} 
            %%{Browse "Die Kante "#@LastEdge#" wurde ausgewhlt!"}
            DaVinciManager,readloop
         else % {Browse @LastData}
            DaVinciManager,readloop
         end
      end
      */   
   end
   
   class DaVinciGraph from DaVinciManager
                         
      meth init(Listener)
         S in
         DaVinciClass, init(S)
         thread
            {ForAll S proc {$ Msg}
                         {Listener Msg}
                         if Msg==quit then
                            {Thread.terminate {Thread.this}}
                         end
                      end}
         end
      end
      meth show(GraphSpec)
         case GraphSpec
         of graph(nodes:Nodes edges:Edges) then
            G = {MakeGraph {MakeChildrenRec Nodes Edges}} in
            DaVinciManager, send(graph(new(G)))
         [] children(...) then
            G = {MakeGraph GraphSpec} in
            DaVinciManager, send(graph(new(G)))
         end
      end
   end
   
end





