%%
%% directory utilities
%%

functor

import
%  System
   OS(getDir stat)
   Open(file)

export
   complete:   Complete
   getFiles:   GetFiles
   getSubdirs: GetSubdirs
   
   File % class File
        %    meth init(PathVS NameVS)
        %    meth create(PathVS NameVS)
        %    meth get(path:     Path     <= _    % --> VirtualString
        %             name:     Name     <= _    % --> VirtualString
        %             stat:     Stat     <= _    % --> stat(mtime: size: type:)
        %             mtime:    MTime    <= _    % --> Integer
        %             size:     Size     <= _    % --> Integer
        %             type:     Type     <= _    % --> Atom, e.g. 'reg'
        %             fullname: Fullname <= _    % --> VirtualString
        %             ?Record            <= _)   % --> file(name: path: stat:)
        %    meth read($ len:?Len<=_)  --> String
        %    meth rewrite(VS ?Ok<=true)
        %    meth append(VS ?Ok<=true)
        %    meth check(size:Size<=@stat.size $<=true)
   
   Dir  % class Dir
        %    meth initTree(PathVS)
        %    meth init(PathVS recursive:Recursive<=false)
        %    meth file(NameVS $)   --> FileObject
        %    meth fileNames($)     --> List of VirtualStrings
        %    meth files($)         --> List of FileObjects
        %    meth checkFiles($<=true)
        %    meth getFiles($)      --> List of File-Headers, see class File
        %    meth subdir(NameVS $) --> DirObject
        %    meth subdirNames($)   --> List of VirtualStrings
        %    meth subdirs($)       --> List of DirObjects
        %    meth checkSubdirs($<=true recursive:Recursive<=true)
        %    meth getSubdirs($)    --> List of dir(path: files: subdirs: stat:)
        %    meth check($<=true recursive:Recursive<=true)
        %    meth get(path:     Path     <= _    % --> VirtualString 
        %             files:    Files    <= unit % --> List of File-Headers
        %             subdirs:  Subdirs  <= unit % --> List of Dir-Headers
        %             stat:     Stat     <= _    % --> stat(mtime: size: type:)
        %             mtime:    MTime    <= _    % --> Integer
        %             size:     Size     <= _    % --> Integer
        %             type:     Type     <= _    % --> should be 'dir'
        %             ?Record            <= unit)% --> Dir-Header
        %    meth getNoSubdirs($)  --> DirHeader
define

   VS2S = VirtualString.toString

   fun {Complete DirVS}
      DirS = {VS2S DirVS} in
      if {List.last DirS} == &/
      then DirS
      else DirS#"/"
      end
   end
   
   fun {GetEntries DirName}
      {Filter {OS.getDir DirName}
       fun {$ C|_} C\=&. end}
   end
   fun {GetFiles DirName}
      {Filter {GetEntries DirName}
       fun {$ Entry} {OS.stat {Complete DirName}#Entry}.type==reg end}
   end
   fun {GetSubdirs DirName}
      {Filter {GetEntries DirName}
       fun {$ Entry} {OS.stat {Complete DirName}#Entry}.type==dir end}
   end
   
   %% from file OzScanner.ozg by Leif Kornstaedt:
   
   %%--------------------------------------------------------------------
   %% File Name Handling

%  PathSeparator = if Windows then &; else &: end

%   fun {IsAbsoluteFileName S}
%      case S of &/|_ then true
%      elseof _|&:|_ then Windows   % good old DOS filename like E:...
%      else false
%      end
%   end

%   %% HACK by afranke
%   GetUserHome    = fun {$ User} "/home/afranke" end
%   IsReadableFile = fun {$ FileName}
%                       URL = {System.property.get url}
%                    in
%                       try
%                          FileA = {URL.localize FileName}
%                       in
%                          if {IsFree FileA}
%                          then false
%                          else true
%                          end
%                       catch _ then
%                          false
%                       end
%                    end
%   %% END OF HACK
   
%%      ozc(getUserHome: GetUserHome isReadableFile: IsReadableFile) =
%%      {Foreign.require 'ozc.o' ozc(getUserHome: 2 isReadableFile: 2)}
   
%   fun {CheckAccess FileName}
%      if {IsReadableFile FileName} then FileName
%      else FileName2 = {Append FileName ".oz"} in
%         if {IsReadableFile FileName2} then FileName2
%         else ""
%         end
%      end
%   end
   
%   fun {GetDirName FileName}
%      {Reverse {List.dropWhile {Reverse FileName} fun {$ C} C \= &/ end}}
%   end
   
%   fun {SearchPath Path FileName}
%      case Path of _|_ then P1 Pr in
%         {List.takeDropWhile Path fun {$ C} C \= PathSeparator end ?P1 ?Pr}
%         case {CheckAccess {Append P1 &/|FileName}} of Res=_|_ then Res
%         elsecase Pr of !PathSeparator|Prr then {SearchPath Prr FileName}
%         [] nil then ""
%         end
%      [] nil then {CheckAccess &.|&/|FileName}
%      end
%   end

%   fun {ExpandFileName FileName CurrentFileName}
%      if {IsAbsoluteFileName FileName} then {CheckAccess FileName}
%      else
%         case FileName of &~|Sr then FileRest UserHome in   % expand `~'
%            case Sr of &/|Srr then
%               FileRest = Srr
%               UserHome = {OS.getEnv "HOME"}
%            else User in
%               FileRest = {List.takeDropWhile Sr fun {$ C} C \= &/ end ?User}
%               UserHome = {GetUserHome User}
%            end
%            case UserHome of "" then ""
%            [] _|_ then {CheckAccess {Append UserHome &/|FileRest}}
%            end
%         else
%            case {CheckAccess {Append {GetDirName CurrentFileName} FileName}}
%            of S2=_|_ then S2
%            [] "" then {SearchPath {OS.getEnv "OZPATH"} FileName}
%            end
%         end
%      end
%   end

   VS2A = VirtualString.toAtom
   
   class File

      attr
         fullname:nil
         path:nil
         name:nil
         stat:nil
      
      meth init(Path Name)
         File,SetFullname(Path Name)
         File,SetStat
      end

      meth SetFullname(Path Name)
         fullname <- {VirtualString.toString Path#"/"#Name}
         File,SetPath(Path)
         File,SetName(Name)
      end
      
      meth SetPath(Path)
         path <- {VirtualString.toString Path}
      end
           
      meth SetName(Name)
         name <- {VirtualString.toString Name}
      end

      meth SetStat
         stat <- {OS.stat @fullname}
      end

      meth create(Path Name)
         F = {New Open.file init(name:  Path#"/"#Name
                                 flags: [write create exclude])}
      in
         {F close}
         File,init(Path Name)
      end
      
      meth get(path:     Path     <= _
               name:     Name     <= _
               stat:     Stat     <= _
               mtime:    MTime    <= _
               size:     Size     <= _
               type:     Type     <= _
               fullname: Fullname <= _
               ?Record            <= _)
         
         Path     = @path
         Name     = @name
         Stat     = @stat
         MTime    = Stat.mtime
         Size     = Stat.size
         Type     = Stat.type
         Fullname = {VirtualString.toString Path#'/'#Name}
         
         Record   = file(path: {StringToAtom @path}
                         name: {StringToAtom @name}
                         stat: @stat)
      end
      
      meth read(?Content len:?Len <= _)
         F = {New Open.file init(name: @fullname)}
      in
         {F read(list:Content size:all len:?Len)}
         {F close}
      end

      meth rewrite(VS ?Ok <= true)
         F = {New Open.file init(name: @fullname
                                 flags: [write truncate])}
      in
         {F write(vs:VS)}
         {F close}
         local
            Len = {Length {VirtualString.toString VS}} in
            File, check(size:Len ?Ok)
         end
         File,SetStat
      end

      meth append(VS ?Ok <= true)
         F = {New Open.file init(name:  @fullname
                                 flags: [write append])}
      in
         {F write(vs:VS)}
         {F close}
         local
            Len = {Length {VirtualString.toString VS}} in
            File, check(size:@stat.size+Len ?Ok)
         end
         File,SetStat
      end
      
      meth check(size:Size <= @stat.size $ <= true)
         File,read(_ len:$) == Size
      end

   end % of class File

   class Dir
      
      attr
         path:"."
         files:nil
         subdirs:nil
         stat:unit

         recursive:false
              
      meth initTree(Path)
         Dir,init(Path recursive:true)
      end              
      
      meth init(Path recursive:Recursive <= false)
         recursive <- Recursive
         Dir,SetPath(Path)
         Dir,SetStat
         @stat.type = dir
         files   <- {NewDictionary}
         subdirs <- {NewDictionary}
         proc {DoInit}
            {ForAll {OS.getDir Path}
             proc {$ EntryS}
                if EntryS == "." orelse EntryS == ".."
                then skip
                else
                   EntryA = {VS2A EntryS}
                in
                   case {OS.stat @path#"/"#EntryS}.type
                   of dir then
                      if @recursive
                      then
                         SubPath = @path#"/"#EntryS
                         DirObj = {New Dir init(SubPath recursive:true)} in
                         {Dictionary.put @subdirs EntryA DirObj}
                      else
                         {Dictionary.put @subdirs EntryA unit}
                      end
                   else         
                      FileObj = {New File init(@path EntryS)} in
                      {Dictionary.put @files EntryA FileObj}               
                   end
                end
             end}
         end
      in
         try
            {DoInit}
         catch _ then
            {Delay 4000}
            Dir, init(Path recursive:Recursive)
         end
      end
      
      meth SetPath(Path)
         path <- {VirtualString.toString Path}
      end
      
      meth SetStat
         stat <- {OS.stat @path}
      end

      meth file(NameVS $)
         NameA = {VS2A NameVS}
      in
         case {Dictionary.condGet @files NameA false}
         of false then raise dir('No such file.' path:@path name:NameVS) end
         elseof FileObj then FileObj
         end
      end
   
      meth fileNames($)
         {Dictionary.keys @files}
      end
      
      meth files($)
         {Dictionary.items @files}
      end
   
      meth checkFiles($ <= true)
         {All Dir,files($) fun {$ FileObj} {FileObj check($)} end}
      end
      
      meth getFiles($)
         {Map Dir,files($) fun {$ FileObj} {FileObj get($)} end}
      end
      
      meth subdir(NameVS ?DirObj)
         NameA = {VS2A NameVS} in
         case {Dictionary.condGet @subdirs NameA false}
         of false then
            raise dir('No such directory.' path:@path dir:NameVS) end
         [] unit then
            DirObj = {New Dir init(@path#"/"#NameVS
                                   recursive: @recursive)}
            {Dictionary.put @subdirs NameA DirObj}
         elseof D then
            D = DirObj
         end
      end
   
      meth subdirNames($)
         {Dictionary.keys @subdirs}
      end
      
      meth subdirs($)
         {Map Dir,subdirNames($) fun {$ Name} Dir,subdir(Name $) end}
      end
  
      meth checkSubdirs($ <= true recursive:Recursive <= true)
         if Recursive orelse @recursive 
         then {All Dir,subdirs($)
               fun {$ DirObj}
                  {DirObj check($ recursive:true)}
               end}
         else {All Dir,subdirNames($)
               fun {$ NameA}
                  case {Dictionary.get @subdirs NameA}
                  of unit then true
                  elseof DirObj then {DirObj check($ recursive:false)}
                  end
               end}
         end
      end
          
      meth getSubdirs($)
         {Map Dir,subdirNames($)
          fun {$ NameA}
             case {Dictionary.get @subdirs NameA}
             of unit then dir(path:{VS2A @path#"/"#NameA}
                              /* stat:_ files:_ subdirs:_ */)
             elseof DirObj then {DirObj get($)}
             end
          end}
      end
      
      meth check($ <= true recursive:Recursive <= true)
         Dir,checkFiles($) andthen Dir,checkSubdirs($ recursive:Recursive)
      end
      
      meth get(path: Path <= _
               files:    Files    <= unit
               subdirs:  Subdirs  <= unit
               stat:     Stat     <= _
               mtime:    MTime    <= _
               size:     Size     <= _
               type:     Type     <= _
               ?Record            <= unit)

         Path  = @path
         Stat  = @stat
         MTime = Stat.mtime
         Size  = Stat.size
         Type  = Stat.type
         
         if {Not {IsDet Files}} orelse Files \= unit
         then Files = Dir,getFiles($)
         else skip
         end
         if {Not {IsDet Subdirs}} orelse Subdirs \= unit
         then Subdirs = Dir,getSubdirs($)
         else skip
         end
         if {Not {IsDet Record}} orelse Record \= unit
         then Record = dir(path:    {StringToAtom @path}
                           stat:    @stat
                           files:   Dir,getFiles($)
                           subdirs: Dir,getSubdirs($))
         else skip
         end
      end
      
      meth getNoSubdirs($)
         dir(path:    {StringToAtom @path}
             stat:    @stat
             files:   Dir,getFiles($)
%            subdirs: _
            )
      end
      
   end % of class Dir
   
end %of functor

