%%%
%%% file utilities
%%%

functor
   
import
   Open(file text)
   OS(stat system unlink)

export
   appendVS: AppendVS %% proc {$ VirtualString FileName}  
   saveVS:   SaveVS   %% proc {$ VirtualString FileName}
   loadVS:   LoadVS   %% fun {$ FileName} --> VirtualString
   loadBS:   LoadBS   %% fun {$ FileName} --> ByteString
   loadS:    LoadS    %% fun {$ FileName} --> String
   
   writeLines:   WriteLines
   writeLinesVS: WriteLines  %% proc {$ VirtualStrings FileName}
   readLines:    ReadLines   
   readLinesVS:  ReadLines   %% fun {$ FileName} --> list of VirtualStrings
   readLinesBS:  ReadLinesBS %% fun {$ FileName} --> list of ByteStrings
   readLinesS:   ReadLines   %% fun {$ FileName} --> list of Strings
   
   exists:      Exists           %% fun {$ FileName} --> Bool
   type:        FileType         %% fun {$ FileName} --> {reg,dir,chr,...}
   backup:      BackupFile       %% proc {$ FileName}
   remove:      RemoveFile       %% proc {$ FileName}
   removeForce: RemoveFileForce  %% proc {$ FileName}
   removeDir:   RemoveDir        %% proc {$ DirName}

define

   %%
   %% load and save
   %%

   proc {AppendVS VS File}
      F = {New Open.file init(name:File flags:[write create append])}
   in
      {F write(vs:VS)}
      {F close}
   end
   
   proc {SaveVS VS File}
      F = {New Open.file init(name:File flags:[write create truncate])}
   in
      {F write(vs:VS)}
      {F close}
   end

   proc {LoadS File ?S}
      F = {New Open.file init(name:File)}
   in
      {F read(list:?S size:all)}
      {F close}
   end
   
   fun {LoadBS File}
      {ByteString.make {LoadS File}}
   end
   
   LoadVS = LoadS

   %%
   %% existence, type, backup and removal
   %%
   
   fun {Exists File}
      try {OS.stat File _} true
      catch _ then false end
   end

   fun {FileType File}
      try {OS.stat File}.type
      catch _ then false end
   end

   %% backup file --> file~ (if not exists: skip) 
   proc {BackupFile File}
      case {FileType File}
      of reg   then {SaveVS {LoadVS File} File#'~'}
      [] false then skip
      else raise backup('File is not regular.' file:File) end
      end
   end

   %% remove file (if not exists: skip)
   proc {RemoveFile File}
      case {FileType File}
      of reg   then {OS.unlink File}
      [] false then skip
      else raise remove('File is not regular.' file:File) end
      end
   end
   
   proc {RemoveFileForce File}
      try {OS.unlink File}
      catch _ then skip end
   end

   proc {RemoveDir Dir}
      case {FileType Dir}
      of false then skip
      [] dir then {OS.system '\\rm -rf '#Dir _}
      else raise remove('Not a directory.' file:Dir) end
      end
   end

   %%
   %% conversion between files and lists of lines
   %%

   class TextFile from Open.text Open.file end

   VS2BS = ByteString.make
   
   proc {WriteLines Lines FileName}
      F = {New TextFile init(name:FileName flags:[read write create truncate])}
   in
      {ForAll Lines
       proc {$ Line} {F putS(Line)} end}
      {F close}
   end
   
   proc {ReadLines FileName ?Lines}
      F = {New TextFile init(name:FileName)}
      fun {Read}
         case {F getS($)} of false then nil
         [] Line then Line|{Read} end
      end
   in
      Lines = {Read}
      {F close}
   end
   
   proc {ReadLinesBS FileName ?Lines}
      F = {New TextFile init(name:FileName)}
      fun {Read}
         case {F getS($)} of false then nil
         [] Line then {VS2BS Line}|{Read} end
      end
   in
      Lines = {Read}
      {F close}
   end
end




