%%%
%%% GDB wrapper class
%%%

functor
   
import
   Error(registerFormatter)
   OS(system)
   Gdbm at 'x-oz://contrib/gdbm'
   
export
   'class': GdbFile
   
define   
   
   VS2BS = ByteString.make

   %%
   %% error formatter
   %%
   {Error.registerFormatter database
    fun {$ E}
       T = 'database error' 
    in 
       case E of database(db(Msg...)) then
          Ar = {List.subtract {Arity E.1} 1}
       in
          error(kind:T msg:Msg
                items: {Map Ar fun {$ F} hint(l:F m:oz(E.1.F)) end})
       else
          error(kind:T items:[line(oz(E))])
       end 
    end}

   proc {GDB_ERROR NameA Obj Method}
      raise database(db(NameA object:{Obj describe($)} method:Method)) end
   end

   %%
   %% simple gdbm object extension
   %%
   class GdbFile
      
      attr
         filename: ''
         gdb:      unit
         mode:     closed
         
      meth describe($)
         gdbFile(filename:@filename gdb:@gdb mode:@mode)
      end
      meth isOpen($)
         @mode \= closed
      end
      meth isClosed($)
         @mode == closed
      end
      meth isWriteable($)
         @mode == write
      end
      meth filename($)
         @filename
      end
      
      meth checkClosed(M)
         if GdbFile, isClosed($) then skip
         else
            {GDB_ERROR db_not_closed self M}
         end
      end
      
      %% the db entity
      meth openCreate(FileName
                      mode:      Mode      <= unit
                      overwrite: Overwrite <= false) = M
         CreateMode = if Overwrite then new else create end
      in
         GdbFile, checkClosed(M)
         {OS.system 'mkdir -p `dirname '#FileName#'`' _}
         gdb      <- if Mode == unit
                     then {Gdbm.new CreateMode(FileName)}
                     else {Gdbm.new CreateMode(FileName mode:Mode)}
                     end
         mode     <- write
         %%{Registry register(self)}
         filename <- {VS2BS FileName}
      end
      meth openWrite(FileName) = M
         GdbFile, checkClosed(M)
         gdb      <- {Gdbm.new write(FileName)}
         mode     <- write
         %%{Registry register(self)}
         filename <- {VS2BS FileName}
      end
      meth openRead(FileName) = M
         GdbFile, checkClosed(M)
         gdb      <- {Gdbm.new read(FileName)}
         mode     <- read
         filename <- {VS2BS FileName}
      end             
      meth close
         {Gdbm.close @gdb}
         %%if (@mode == write) then
         %%   {Registry unregister(self)}
         %%end
         mode <- closed
      end
      meth reorganize
         {Gdbm.reorganize @gdb}
      end
      
      %% single entry
      meth member(Key $)
         {Gdbm.member @gdb Key}
      end
      meth get(Key $)
         {Gdbm.get @gdb Key}
      end
      meth condGet(Key Default $)
         {Gdbm.condGet @gdb Key Default}
      end
      meth put(Key Val) = M
         {Gdbm.put @gdb Key Val}
      end
      meth remove(Key) = M
         {Gdbm.remove @gdb Key}
      end
      
      %% all entries
      meth items($)
         {Gdbm.items @gdb}
      end
      meth keys($)
         {Gdbm.keys @gdb}
      end
      meth entries($)
         {Gdbm.entries @gdb}
      end
      meth forAll(P)
         {Gdbm.forAll @gdb P}
      end
      meth forAllInd(P)
         {Gdbm.forAllInd @gdb P}
      end
   end
      
end





