%%%
%%% internet tools
%%%

functor

import
   %%Error
   OS(getEnv getHostByName uName tmpnam system unlink)
   Property(get)

   File(loadS)
   at 'x-ozlib://mathweb/share/File.ozf'
   
export
   readIP:     ReadIP
   writeIP:    WriteIP
   getIP:      GetIP
   isIP:       IsFullAddress

   splitAddress:  SplitAtDots
   concatAddress: ConcatDots

   sameHosts:  SameHosts
   
   nodename:   NODENAME
   hostname:   HOSTNAME
   address:    ADDRESS
   
   expand:     ExpandHostname

   display:    DISPLAY
   
define

   VS2S = VirtualString.toString

   IsPrefix = List.isPrefix
   
   %%
   %% split an internet address into its components
   %%
   %% examples: "leibniz.ags.uni-sb.de" --> ["leibniz" "ags" "uni-sb" "de"]
   %%            "134.96.236.125"       --> ["134" "96" "236" "125"]
   local
      proc {SplitAtDot S ?S1 ?S2}
         case S of nil then S1=nil S2=false
         [] C|Sr then
            if C==&.
            then S1=nil S2=Sr
            else S1=C|{SplitAtDot Sr $ ?S2}
            end
         end
      end
   in
      fun {SplitAtDots S}
         S1 S2 in
         {SplitAtDot S ?S1 ?S2}
         if S2==false then [S1]
         else S1|{SplitAtDots S2} end
      end
   end

   %%
   %% concatenate the address components again to make a virtual string
   %%
   local
      fun {DottedTail Xs}
         case Xs of nil then ""
         [] X|Xr then '.'#X#{DottedTail Xr} end
      end
   in
      fun {ConcatDots Parts}
         First|Rest = Parts in
         First#{DottedTail Rest}
      end
   end
   
   %%
   %% convert an IP address string into a tuple
   %%
   %% example: "134.96.236.125" --> '#'(134 96 236 125)
   %%                             =     134#96#236#125
   fun {ReadIP S}
      {List.toTuple '#' {Map {SplitAtDots S} StringToInt}}
   end
   
   %%
   %% convert an IP address tuple into a virtual string 
   %%
   fun {WriteIP IP}
      IP.1#'.'#IP.2#'.'#IP.3#'.'#IP.4
      %% alternatively:
      %% {ConcatDots {Record.toList IP}}
   end
       
   %%
   %% get the first element of the host's IP address list 
   %%   
   proc {GetIP Host ?S}
      S|_ = {OS.getHostByName Host}.addrList 
   end

   %%
   %% recognize IP addresses
   %%
   fun {IsFullAddress S}
      {IsString S} andthen
      local
         Parts = {SplitAtDots S} in
         {Length Parts} == 4 andthen
         {All Parts String.isInt}
      end
   end

   %%
   %% recognize fully qualified hostnames
   %%
   fun {IsFullName S}
      Parts = {SplitAtDots S} in
      {Length Parts} > 1
   end
   
   %%
   %% host name comparison
   %%
   local
      fun {GetList Host}
         try
            H = {OS.getHostByName Host} in
            H.name|{Append H.aliases H.addrList}
         catch _ then nil end
      end
   in
      fun {SameHosts Host1 Host2}
         {IsDet Host1} andthen {IsVirtualString Host1} andthen
         {IsDet Host2} andthen {IsVirtualString Host2} andthen
         local
            H1 = {GetList Host1} 
            H2 = {GetList Host2}
         in
            {Some H1 fun {$ X} {Member X H2} end}
         end
      end
   end

   /*
   %%
   %% validate a hostname
   %%
   fun {IsValidName N}
      try _={OS.getHostByName N} true
      catch _ then false end
   end
   */
   
   %% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   %% the easy part: use the Oz-builtins 
   %% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

   NODENAME   = {OS.uName}.nodename
   H          = {OS.getHostByName NODENAME}

   FullNames
   SimpleNames
   {List.partition H.name|H.aliases IsFullName ?FullNames ?SimpleNames}

   NonTrivNames = {Filter SimpleNames fun {$ N} N\="localhost" end}
   NonTrivAddrs = {Filter H.addrList  fun {$ A} A\="127.0.0.1" end}
   
   /*
   NODENAME = if Nodename == "localhost" then      
                 case {Filter SimpleNames fun {$ N} N\="localhost" end}
                 of nil then
                    case FullNames of nil then "localhost"
                    [] F1|_ then {List.nth {SplitAtDots F1} 1}
                    end
                 [] N1|_ then N1
                 end
              else Nodename end
   */

   /*
   %%
   %% Failure
   %%
   local
      proc {RegisterFormatter}
         {Error.registerFormatter internet
          fun {$ E}
             Title     = 'hostname problem'
             BugReport = 'please send a bug report to afranke@ags.uni-sb.de'
          in
             case E of internet(internal(X)...) then 
                error(kind:Title msg:'Internal error'
                      items: [hint(l:'Additional information' m:oz(X))
                              unit line(BugReport)])
             [] internet(fullname(Msg...)...) then
                error(kind:Title msg:Msg
                      items: {Append
                              {Map {List.subtract {Arity E.1} 1}
                               fun {$ F} hint(l:F m:oz(E.1.F)) end}
                              [unit line(BugReport)]})
             else error(kind:Title items:[line(oz(E))])
             end 
          end}
      end
      fun {StdInfo}
         VS2A = VirtualString.toAtom in
         info(name:     {VS2A H.name}
              aliases:  {Map H.aliases VS2A}
              addrList: {Map H.addrList VS2A}
              platform: {Property.get 'platform.name'}
              path:     case {OS.getEnv "PATH"} of false then false
                        [] PATH then {VS2A PATH} end)
      end
      proc {GenericFail ErrRec}
         {RegisterFormatter}
         raise internet({Adjoin {StdInfo} ErrRec}) end
      end
   in
      proc {Fail FailRec} 
         case {Label FailRec}
         of fullname then
            {GenericFail {AdjoinAt FailRec 1
                          'Could not determine fully qualified hostname'}}
         else
            {GenericFail internal(FailRec)}
         end    
      end
   end
   */
   
   %%
   %% try to find out the (dns-)domainname with help of unix commands
   %%
   local
      fun {FirstWord S}
         {List.takeWhile {List.dropWhile S Char.isSpace} 
          fun {$ C} {Char.isSpace C}==false end}
      end
      proc {SystemOutput Cmd ?Output} % may return nil
         Tmp = {OS.tmpnam}
      in
         if {OS.system Cmd#" > "#Tmp#" 2>/dev/null" $} == 0
         then {File.loadS Tmp ?Output}
         else Output=nil
         end
         try {OS.unlink Tmp}
         catch system(os(os "unlink" 2 _)...) then skip end
      end
      fun {FirstOutput Cmd}
         {FirstWord {SystemOutput Cmd}}
      end
      fun {Which Cmd} % may return nil
         Cmd1={FirstWord Cmd}
         CmdR={List.drop Cmd {Length Cmd1}}
      in
         case {FirstOutput "which "#Cmd1}
         of nil  then nil
         [] "no" then nil
         [] FullCmd1 then {Append FullCmd1 CmdR}
         end
      end
      fun {FindCmd PrefixList Cmd}
         case PrefixList of nil then {VS2S "NOTFOUND_"#Cmd}
         [] L1|Lr then
            case {Which {VS2S L1#Cmd}}
            of nil then {FindCmd Lr Cmd}
            elseof Result then Result end 
         end
      end
      fun {FindByNeed PrefixList Cmd}
         {Value.byNeed fun {$} {FindCmd ""|PrefixList Cmd} end}
      end
   in
      fun {CmdFound CmdVar}
         {List.isPrefix "NOTFOUND_" CmdVar}==false
      end
      fun {CmdsFound Cmds}
         {All Cmds CmdFound}
      end
      
      `nslookup` = {FindByNeed ["/usr/sbin/" "/usr/bin/" "/bin/"
                                "/system/bin/" "/usr/local/bin/"] "nslookup"}
      
      `grep` = {FindByNeed ["/bin/" "/usr/bin/" "/opt/gnu/bin/"] "grep"}
      `awk`  = {FindByNeed ["/bin/" "/usr/bin/" "/opt/gnu/bin/"] "awk"}
      `tail` = {FindByNeed ["/usr/bin/"] "tail"}
      
      `domainname`    = {FindByNeed ["/bin/"] "domainname"}
      `dnsdomainname` = {FindByNeed ["/bin/"] "dnsdomainname"} %% linux only
      `hostname`      = {FindByNeed ["/bin/"] "hostname"}
      
      fun {`_domainname`}    {FirstOutput `domainname`}            end
      fun {`_dnsdomainname`} {FirstOutput `dnsdomainname`}         end
      %%fun {`_hostname -f`}   {FirstOutput {VS2S `hostname`#" -f"}} end
      fun {`_hostname -d`}   {FirstOutput {VS2S `hostname`#" -d"}} end
      %%fun {`_hostname`}      {FirstOutput {VS2S `hostname`}}       end
      fun {`_nslookup Name` Host}
         if {CmdsFound [`nslookup` `grep` `awk`]} then
            {FirstOutput {VS2S `nslookup`#" "#Host#"|"#`grep`#" Name: "
                          #"| "#`awk`#" '{print $2}'"}}
         else nil end
      end
      fun {`_nslookup Address` Host}
         if {CmdsFound [`nslookup` `grep` `tail` `awk`]} then
            {FirstOutput {VS2S `nslookup`#" "#Host#"|"#`grep`#" Address: "
                          #"| "#`tail`#" -1 | "#`awk`#" '{print $2}'"}}
         else nil end
      end
      %% read the 'domain' entry in /etc/resolv.conf
      fun {ResolvConfDomain}
         if {CmdsFound [`grep` `awk`]} then
            {FirstOutput {VS2S `grep`#" domain /etc/resolv.conf "#
                          "| "#`awk`#" '{print $2}'"}}
         else nil end
      end
      %% read the 'search' entry in /etc/resolv.conf
      local
         fun {ReadNth N}
            {FirstOutput {VS2S `grep`#" search /etc/resolv.conf "#
                          "| "#`awk`#" '{print $"#(1+N)#"}'"}}
         end
         fun {ReadRest N}
            case {ReadNth N} of nil then nil
            [] Entry then Entry|{ReadRest N+1} end
         end
      in
         fun {SearchList}
            if {CmdsFound [`grep` `awk`]} then {ReadRest 1}
            else nil end
         end
      end
   end

   %%
   %% FindFullname
   %%
   fun {FindFullName Names}
      case Names of nil then false
      [] N|Nr then
         if {IsFullName N} then true(N)
         else {FindFullName Nr} end
      end
   end

   fun {First Xs Fun}
      case Xs of nil then nil
      [] X|Xr then
         if {Fun X} then X else {First Xr Fun} end
      end
   end
 
   %%
   %% find out fully qualified hostname and IP address with help of nslookup
   %%
   fun {FindHostname Names}
      case Names of nil then nil
      [] N1|Nr then
         case {`_nslookup Name` N1}
         of nil then {FindHostname Nr}
         [] Fullname then Fullname
         end
      end
   end
   
   fun {FindAddress Names}
      case Names of nil then nil
      [] N1|Nr then
         case {`_nslookup Address` N1}
         of nil then {FindAddress Nr}
         [] Addr then Addr
         end
      end
   end

   fun {BestAddress}
      case NonTrivAddrs of A1|_ then A1
      [] nil then
         case {FindAddress {Append FullNames NonTrivNames}}
         of nil then {List.nth H.addrList 1}
         [] Addr then Addr
         end
      end
   end
   
   %% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   %% hostname expansion methods  
   %% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

   FailExpansion = fun {$ N} nil end

   GetHostByNameExpansion = fun {$ N}
                               try
                                  H={OS.getHostByName N} in
                                  {First H.name|H.aliases IsFullName}
                               catch _ then nil end
                            end
   
   NslookupExpansion = {Value.byNeed
                        fun {$}
                           if {CmdsFound [`nslookup` `grep` `awk`]} then
                              fun {$ N}
                                 try {`_nslookup Name` N}
                                 catch _ then nil end
                              end
                           else FailExpansion end
                        end}

   SearchExpansion = local
                        fun {Lookup N L}
                           case L of nil then nil
                           [] L1|Lr then
                              try
                                 _={OS.getHostByName N#'.'#L1} in
                                 {VS2S N#'.'#L1}
                              catch _ then {Lookup N Lr} end
                           end
                        end
                     in
                        {Value.byNeed
                         fun {$} 
                            case {SearchList}
                            of nil then FailExpansion
                            [] SL  then fun {$ N} {Lookup N SL} end
                            end
                         end}
                     end

   NonTrivAddrExpansion = fun {$ N}
                             try
                                case {OS.getHostByName N}.addrList
                                of nil then nil
                                [] "127.0.0.1"|NonTriv then
                                   case NonTriv
                                   of nil then nil
                                   [] A|_ then A end
                                [] A|_ then A end
                             catch _ then nil end
                          end

   AdditionalDomainsExpansion =
   local
      fun {Lookup N L}
         case L of nil then nil
         [] nil|Lr then {Lookup N Lr} 
         [] L1|Lr then
            try
               Hname = {VS2S N#'.'#L1}
               _     = {List.nth {OS.getHostByName Hname}.addrList 1}
            in Hname
            catch _ then {Lookup N Lr} end
         end
      end
   in
      fun {$ N}
         {Lookup N [{Value.byNeed ResolvConfDomain}
                    {Value.byNeed `_dnsdomainname`}
                    {Value.byNeed `_hostname -d`}
                    {Value.byNeed `_domainname`}
                   ]}
      end
   end

   ExpandHostname = 
   local
      fun {Try Methods N}
         case Methods of nil then N
         [] false|Mr then {Try Mr N}
         [] M|Mr then
            case {M N} of nil then {Try Mr N}
            [] Expanded then Expanded
            end
         end
      end
   in
      case {Property.get 'platform.os'}
      of win32 then
         fun {$ N} N end
      [] linux then
         fun {$ N}
            {Try [GetHostByNameExpansion
                  NslookupExpansion
                  SearchExpansion
                  AdditionalDomainsExpansion
                  NonTrivAddrExpansion] N}
         end
      [] solaris then
         fun {$ N}
            {Try [GetHostByNameExpansion
                  NslookupExpansion
                  SearchExpansion
                  AdditionalDomainsExpansion
                  NonTrivAddrExpansion] N}
         end
      else
         fun {$ N} N end
      end
   end
   
   %% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   %% set ADDRESS and HOSTNAME (avoid "127.0.0.1" and "localhost", if possible)
   %% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

   ADDRESS
   HOSTNAME
   
   if {IsFullName H.name} then
      HOSTNAME   = H.name
      ADDRESS|_  = H.addrList 
   elseif FullNames\=nil then
      HOSTNAME = case {First FullNames fun {$ N} {IsPrefix NODENAME N} end}
                 of false then
                    case {First FullNames fun {$ N} {IsPrefix H.name N} end}
                    of false then
                       {List.nth FullNames 1} 
                    elseof N then N end
                 elseof N then N end
      ADDRESS = {BestAddress}
   elseif NonTrivNames\=nil then
      HOSTNAME = case {FindHostname NonTrivNames} of nil then
                    {ExpandHostname {List.nth NonTrivNames 1}}
                 elseof N then N end
      ADDRESS  = {BestAddress}
   else
      {ExpandHostname NODENAME}
      ADDRESS = {BestAddress}     
   end
   /*
   case ADDRESS of "127.0.0.1" then
      HOSTNAME = NODENAME
   elsecase {FindFullName H.name|H.aliases} of true(FullName) then
      HOSTNAME = FullName
   elsecase {Property.get 'platform.os'}
   of win32 then {Fail fullname}
   [] linux then HOSTNAME {UseFallback}
   [] solaris then
      Hostname = {`_nslookup Name` NODENAME} in
      HOSTNAME = if {IsFullName Hostname}
                 then Hostname
                 else {UseFallback}
                 end
   else
      HOSTNAME = {UseFallback}
   end
   */
   
   %%
   %% $DISPLAY variable with expanded Hostname
   %%
   local
      proc {SplitAtColon S ?S1 ?S2}
         case S of nil then S1=nil S2=false
         [] C|Sr then
            if C==&:
            then S1=nil S2=Sr
            else S1=C|{SplitAtColon Sr $ ?S2}
            end
         end
      end
   in
      fun {UnixDisplay}
         case {OS.getEnv "DISPLAY"} of false then false
         [] Display then
            Host No in
            {SplitAtColon Display ?Host ?No}
            if No==false then nil % or false?
            elseif Host == nil orelse Host == NODENAME
            then {VS2S HOSTNAME#':'#No}
            else {VS2S {ExpandHostname Host}#':'#No}
            end
         end
      end
   end
   
   DISPLAY = case {Property.get 'platform.os'}
             of win32   then false
             [] linux   then {UnixDisplay}
             [] solaris then {UnixDisplay}
             else {UnixDisplay}
             end
end
  
%%
%% auxiliuary functions
%%
/*
fun {Find SubS S}
   if {List.isPrefix SubS S} then S
   elsecase S of nil then false
   [] _|Sr then {Find SubS Sr} end
end
*/

%%
%% get the domainname from a full hostname by dropping the first component
%%
/*
fun {DropNodename FullHostname}
   _|R = {SplitAtDots FullHostname} in
   {VS2S {ConcatDots R}}
end
*/ 

/*   
%% Fallback method
fun {Try L}
   case L of nil then nil
   [] L1|Lr then
      case {L1} of nil then {Try Lr}
      elseof Result then Result end 
   end
end

fun {`nslookupdomain`}
   case {`_nslookup Name` NODENAME} of nil then nil
   elseof Hostname then
      if {IsFullAddress Hostname} then {DropNodename Hostname}
      else nil end
   end
end

proc {UseFallback ?Hostname}
   Hostname = if {IsFullName H.name} then H.name
              else
                 case {`_hostname -f`} of nil then ADDRESS
                 elseof Fullname then Fullname end
              end
end
{Value.byNeed fun {$} {DropNodename {`_hostname -f`}} end}
*/


