%%
%% misc
%%

functor

import
   Open OS Property(get) Tk TkTools
   
   XCompiler
   at 'x-ozlib://mathweb/share/XCompiler.ozf'
   
export
   pathSep             : PathSep
   completeDir         : CompleteDir
   mkDir               : MkDir
   checkSize           : CheckSize
   stringLess          : StringLess
   commandToTitle      : CommandToTitle
   atomsToStrings      : AtomsToStrings
   writeLog            : WriteLog
   isFeature           : IsFeature
   errorMessage        : ErrorMessage
   lookup              : Lookup
   %%MenuManager functions:
   printRepresentation : PrintRepresentation
   insertEntry         : InsertEntry
   doEntries           : DoEntries
   makeClass           : MakeClass
   getFeatures         : GetFeatures
   makeKey             : MakeKey
   makeMessage         : MakeMessage
   makeMenu            : MakeMenu
   makePopupMenu       : MakePopupMenu
   virtualStringToValue: VirtualStringToValue

   mapMenu: MapMenu
   
define

   %%
   %% Paths
   %%
   local
      OS      = {Property.get 'platform.os'} in
      WINDOWS = {List.isPrefix "win" {AtomToString OS}}
   end

   PathSep = if WINDOWS
             then &\\
             else &/
             end
   
   fun {CompleteDir DirVS}
      DirS = {VirtualString.toString DirVS} in
      if {Member {List.last DirS} [&/ &\\]} then DirS else DirS#[PathSep] end
   end

   proc {MkDir PathVS}
      MKDIR = if WINDOWS
              then "mkdir"
              else "\\mkdir -p"
              end
   in
      case {OS.system MKDIR#" "#PathVS} of 0 then skip
      [] ErrCode then raise mkdir(ErrCode) end
      end
   end
   
   %%
   %% a 'Compiler.virtualStringToValue' clone
   %%
   VirtualStringToValue = {XCompiler.new}

   %%
   %% ???
   %%
   local
      fun {ComputeCM Is}
         case {Reverse Is}
         of Unit|RevNum then
            Num = {Reverse RevNum}
            N   = if {String.isInt Num} then
                     {Int.toFloat {String.toInt Num}}
                  elseif {String.isFloat Num} then
                     {String.toFloat Num}
                  else false
                  end
            F   = case Unit    
                  of &i then 2.54
                  [] &c then 1.00
                  [] &m then 10.0 % really??? maybe: 0.1 /afranke
                  [] &p then 0.035277778
                  else false
                  end
         in
            if N==false orelse F==false then false
            else N*F
            end
         end
      end      
   in
      fun {CheckSize Is}
         RW XRH
      in
         {List.takeDropWhile {Filter Is Char.isGraph}
          fun {$ I} I\=&x end ?RW ?XRH}
         case XRH of &x|RH then
            case {ComputeCM RW} of false then false
            elseof W then
               case {ComputeCM RH} of false then false
               elseof H then o(width:W height:H)
               end
            end
         else false
         end
      end
   end

   %%
   %% lexicographic ordering on strings
   %%
   %% maybe this is something for a module 'XString',
   %% 'XList' or the like
   %%
   fun {StringLess Is Js}
      case Is of nil then true
      [] I|Ir then
         case Js of nil then false
         [] J|Jr then I<J orelse I==J andthen {StringLess Ir Jr}
         end
      end
   end

   %%
   %% append a line of Text to the logfile in Dir
   %%
   %% maybe something like this can be put into 'File.oz'
   %%
   proc {WriteLog Text Dir}
      try
         File = {New Open.file init(name : Dir#'logfile'
                                    flags: [write create append]
                                    mode : mode(owner:[write] all:[read]))}
      in
         {File write(vs:Text#'\n')}
         {File close}
      catch _ then skip end
   end
   
   %%
   %% type test for features
   %%
   %% maybe this is obsolete because it can be replaced where used
   %%
   fun {IsFeature X}
      {IsInt X} orelse {IsLiteral X}
   end

   %%
   %% show error message in a new window 
   %%
   %% maybe this can be moved to XTkTools
   %%
   proc {ErrorMessage Text}
      {Wait {New TkTools.error tkInit(text:Text aspect:1000)}.tkClosed}
   end

   %%
   %% check first if a value is defined as an emulator property;
   %% if that fails then use the system environment
   %%
   fun {Lookup VarA Default}
      try {Property.get VarA}
      catch system(system(getProperty...)...) then
         case {OS.getEnv VarA} of false then Default
         [] Val then Val end 
      end
   end

   %%
   %% convert a list of atoms into a list of strings
   %%
   %% maybe obsolete, since the same as {Map As AtomToString} ?
   %%
   fun {AtomsToStrings As}
      case As of nil then nil
      [] A|Ar then
         {Atom.toString A}|{AtomsToStrings Ar}
      end
   end
   
   %%
   %% change style of command representation
   %%
   %% same as {PrintRepresentation {AtomToString Command}} ?!
   %%
   fun {CommandToTitle Command}
      fun {CommandToTitleRecursive Cs}
         case Cs of nil then nil
         elseof &-|C|Cr then 32|C|{CommandToTitleRecursive Cr}
         elseof C|Cr then {Char.toLower C}|{CommandToTitleRecursive Cr}
         end
      end
   in
      case {Atom.toString Command} of nil then nil
      elseof C|Cr then
         C|{CommandToTitleRecursive Cr}
      end
   end
   
   %%
   %% "EXAMPLE-STRING" -> "Example string" (?)
   %%
   fun {PrintRepresentation String}
      fun {PrintRepresentationRecursive Cs}
         case Cs of nil then nil
         elseof &-|C|Cr then 32|C|{PrintRepresentationRecursive Cr}
         elseof C|Cr then {Char.toLower C}|{PrintRepresentationRecursive Cr}
         end
      end
   in
      case String of nil then nil
      elseof C|Cr then
         C|{PrintRepresentationRecursive Cr}
      end
   end
   
%%%
%%% maybe the following parts can be separated and moved to
%%% something like 'MenuSupport' together with the rest
%%% of the generic MenuManager stuff
%%%
   
   %%
   %% InsertEntry & DoEntries
   %%
   local
      proc {InsertEntryR Xs E I ?Ys ?P}
         case Xs of nil then Ys=[E] P=I
         [] X|Xr then
            if X==E then % entry already exists
               Ys=X|Xr P=false
            elseif {StringLess X E} then
               Ys=X|{InsertEntryR Xr E I+1 $ P}
            else
               Ys=E|X|Xr P=I
            end
         end
      end
   in
      proc {InsertEntry Xs E ?Ys ?P}
         {InsertEntryR Xs E 0 ?Ys ?P}
      end
   end
   proc {DoEntries ID BD Es W}
      case Es of nil then skip
      [] E|Er then {DoEntries ID BD E W} {DoEntries ID BD Er W}
      else
         if {IsFeature Es} then
            Entry  = {Dictionary.condGet ID Es false}
            Button = {Dictionary.condGet BD Es false}
         in
            if Entry\=false then
               case Entry of En#Ks then
                  {En tk(entryconf state:W)} % change state of menu entry
                  {DoEntries ID BD Ks W}     % and recursively subentries
               else
                  {Entry tk(entryconf state:W)} % change state of menu entry
               end
               if Button\=false then
                  {Button tk(conf state:W)}  % change state of button 
               else
                  skip
               end
            else
               {ErrorMessage 'Unknown menu entry to change state:\n'#Es}
            end
         else
            {ErrorMessage 'Wrong type of menu entry to change state:\n'#Es}
         end
      end
   end

   %%
   %% MakeClass & GetFeatures
   %%
   %% maybe one could create something like 'XClass' ? 
   %%
   fun {MakeClass C Fs}
      case Fs of nil then C else
         {Class.new [C] 'attr' {List.toTuple 'feat' Fs} [final]}
      end
   end
   fun {GetFeatures Ms}
      case Ms of nil then nil
      [] M|Mr then
         if {HasFeature M feature} then M.feature|{GetFeatures Mr}
         else {GetFeatures Mr}
         end
      end
   end

   %%
   %% MakeKey
   %%
   local
      local
         fun {DoMakeEvent R}
            case R
            of ctrl(S) then 'Control-' # {DoMakeEvent S}
            [] alt(S)  then 'Alt-'     # {DoMakeEvent S}
            [] meta(S) then 'Meta-'    # {DoMakeEvent S}
            else R
            end
         end
      in
         fun {MakeEvent R}
            '<' # {DoMakeEvent R} # '>'
         end
      end
   in
      proc {MakeKey M Menu Item KeyBinder}
         if {HasFeature M key} then
            B={HasFeature M event}
            E={MakeEvent M.key}
         in
            {Tk.send bind(KeyBinder
                          if B
                          then M.event
                          else E
                          end
                          v('{') Menu invoke Item v('}'))}
         end
      end
   end

   %%
   %% MakeMessage
   %%
   local
      AccSpace  = '    '
      AccCtrl   = 'Ctrl+'
      AccAlt    = 'Alt+'
      AccMeta   = 'Meta+'
   
      fun {MakeAcc R}
         case R
         of ctrl(S) then AccCtrl # {MakeAcc S}
         [] alt(S)  then AccAlt  # {MakeAcc S}
         [] meta(S) then AccMeta # {MakeAcc S}
         [] less    then '<'
         [] greater then '>'
         else R
         end
      end
   
      proc {ProcessMessage As M ?AMs}
         case As of nil then AMs=nil
         [] A|Ar then AMr in
            AMs = case A
                  of key     then acc#(AccSpace#{MakeAcc M.key})|AMr
                  [] event   then AMr
                  [] feature then AMr
                  [] menu    then AMr
                  else A#M.A|AMr
                  end
            {ProcessMessage Ar M AMr}
         end
      end
   in
      fun {MakeMessage M P}
         {AdjoinList tkInit parent#P|{ProcessMessage {Arity M} M}}
      end
   end

   %%
   %% substitute in menu specifications:
   %%    - objects in Obj#Method actions
   %%    - variables
   %%
   %% also changed: MakePopupMenu /afranke 27.10.99
   local
      fun {IsFeature X}
         {IsLiteral X} orelse {IsInt X}
      end
   in
      fun {MakeAction ObjMap Action}
         if {IsProcedure Action} then Action
         elsecase Action of O#M then
            if {IsFeature O} then (ObjMap.O)#M
            else Action end
         else
            ObjMap.default#Action
         end
      end
      fun {MakeVariable VarMap V}
         VarMap.V
      end
      fun {MapMenu ObjMap VarMap Ms}
         case Ms of nil then nil
         [] M|Mr then
            case {Label M} of separator then separator
            [] cascade then
               {AdjoinAt M menu {MapMenu ObjMap VarMap M.menu}}
            else
               {Record.mapInd M
                fun {$ F V}
                   case F of var then {MakeVariable VarMap V}
                   [] action then {MakeAction ObjMap M.action}
                   else V end
                end}
            end
            | {MapMenu ObjMap VarMap Mr}
         end
      end
   end

   %%
   %% MakePopupMenu
   %%
   local
      proc {MakeMenu ObjMap Parent Ms}
         case Ms of nil then skip
         [] M|Mr then
            case {Label M} of separator then
               _ = {New Tk.menuentry.separator tkInit(parent: Parent)}
            [] cascade then
               CascadeMenu = {MakePopupMenu ObjMap Parent M}
            in
               _ = {New Tk.menuentry.cascade tkInit(parent: Parent
                                                    label : M.label
                                                    menu  : CascadeMenu)}
            else
               Action = {MakeAction ObjMap M.action}
            in
               _ = {New Tk.menuentry.command tkInit(parent: Parent
                                                    label : M.label
                                                    action: Action)}
            end
            {MakeMenu ObjMap Parent Mr}
         end
      end
   in
      fun {MakePopupMenu ObjMap Parent M}
         Popup = {New Tk.menu tkInit(parent : Parent
                                     tearoff: false)}
         {MakeMenu ObjMap Popup M.menu}
      in
         Popup
      end
   end
   
   %%
   %% MakeMenu
   %%
   local
      fun {MakeItems Ms Item Menu KeyBinder Name Dict}
         case Ms of nil then nil
         [] M|Mr then
            HasMenu = {HasFeature M menu}
            BaseCl  = Tk.menuentry.{Label M}
            UseCl   = if HasMenu then
                         FS={GetFeatures M.menu}
                      in
                         {MakeClass BaseCl menu|FS}
                      else BaseCl
                      end
            M1 = {MakeMessage M Menu}
            NewItem = {New UseCl M1}
            Key
            KeyList
         in
            {MakeKey M Menu NewItem KeyBinder}
            if HasMenu then
               NewMenu = {New Tk.menu tkInit(parent:Menu)}
            in
               NewItem.menu = NewMenu
               KeyList = {MakeItems M.menu NewItem NewMenu KeyBinder Name#M.label#'.' Dict}
               {NewItem tk(entryconf menu:NewMenu)}
            end
            if {HasFeature M feature} then
               Item.(M.feature)=NewItem
            end
            if {HasFeature M label} then
               Key = {StringToAtom {VirtualString.toString Name#M.label}}
               if HasMenu
               then {Dictionary.put Dict Key NewItem#KeyList}
               else {Dictionary.put Dict Key NewItem}
               end
            end
            if {IsDet Key} then
               Key|{MakeItems Mr Item Menu KeyBinder Name Dict}
            else
               {MakeItems Mr Item Menu KeyBinder Name Dict}
            end
         end
      end      
      fun {MakeButtons Ms Bar KeyBinder Dict}
         case Ms of nil then nil
         [] M|Mr then
            MenuButton = {New {MakeClass Tk.menubutton
                               menu|{GetFeatures M.menu}}
                          {MakeMessage M Bar}}
            Menu       = {New Tk.menu tkInit(parent:MenuButton)}
         in
            _ = {MakeItems M.menu MenuButton Menu KeyBinder M.text#'.' Dict}
            {MenuButton tk(conf menu:Menu)}
            if {HasFeature M feature} then
               Bar.(M.feature)=MenuButton
            end
            MenuButton.menu = Menu
            MenuButton | {MakeButtons Mr Bar KeyBinder Dict}
         end
      end
   in      
      fun {MakeMenu Parent KeyBinder L R Dict}
         MenuBar      = {New {MakeClass Tk.frame {Append {GetFeatures L}
                                                  {GetFeatures R}}}
                         tkInit(parent:  Parent
                                border: 1%Configure.border
                                'class': 'MenuFrame'
                                relief:  raised)}
         LeftButtons  = {MakeButtons L MenuBar KeyBinder Dict}
         RightButtons = {MakeButtons R MenuBar KeyBinder Dict}
      in
         case {Append
               case LeftButtons of nil then nil
               else [pack(b(LeftButtons) side:left fill:x)]
               end
               case RightButtons of nil then nil
               else [pack(b(RightButtons) side:right fill:x)]
               end}
         of nil then skip
         elseof Tcls then {Tk.batch Tcls}
         end
         MenuBar
      end
   end
end
