%%%
%%% This file provides an abstraction to launch a pipe with an external program
%%% and connect its stdin/stdout streams to an arbitrary remote socket for
%%% message oriented information exchange.
%%%
%%% Once it has been cleaned up, it should be merged back with XOpen.oz
%%%

functor
   
import              
   Error Open(pipe:Pipe) %OS

   Process
   at 'x-ozlib://mathweb/share/Process.ozf'

   XOpen(socket:XSocket)
   at 'x-ozlib://mathweb/share/XOpen.ozf'

   System(show:Browse)
export
   pipe0: Pipe0
   pipe1: Pipe1
   pipe: XPipe
   RunClient
   
define

   fun {`Record.subtractList` R L}
      {Record.filterInd R fun {$ F _} {Member F L}==false end}
   end
         
   proc {Split S StopChar ?S1 ?S2}
      case S of C|Sr then
         if C==StopChar
         then S1=nil S2=Sr
         else S1=C|{Split Sr StopChar $ ?S2}
         end
      [] nil then
         S1 = S2 = S 
         raise connectionClosed end
      end
   end
   proc {Split0 S ?S1 ?S2}
      case S of C|Sr then
         if C==0
         then S1=nil S2=Sr
         else S1=C|{Split0 Sr $ ?S2}
         end
      [] nil then
         S1 = S2 = S
         raise connectionClosed end
      end
   end
   proc {Split128 S ?S1 ?S2}
      case S of C|Sr then
         if C==128
         then S1=nil S2=Sr
         else S1=C|{Split128 Sr $ ?S2}
         end
      [] nil then
         S1 = S2 = S
         raise connectionClosed end
      end
   end

   class Pipe0
      from Pipe
         
      feat
         pid
         readDesc
         
      meth init(cmd:Cmd args:Args pid:?PID<=_) = Init
         self.pid      = Pipe, init(cmd:Cmd args:Args pid:PID=$)
         self.readDesc = Pipe,getDesc($ _)
      end
      
      meth close(kill:Kill<=false)
         %%{OS.deSelect self.readDesc}
         if Kill andthen {IsDet self.pid} then
            %%{OS.system 'kill -9 '#self.pid _}
            {Process.killTree self.pid}
         end
         thread Pipe,close end
         {Delay 600}
      end
   end
   
   class Pipe1
      from Pipe0
      prop locking

      feat
         ReadStopC
         WriteAppendVS
         SplitP
         ReaderThread
         Terminated
         
      attr
         list: unit
         tail: unit
         
      meth init(cmd:Cmd args:Args pid:?PID<=_ ...) = M
         Pipe0, init(cmd:Cmd args:Args pid:?PID)
         RestMethod = {`Record.subtractList` M [cmd args pid]} in
	 {self {Adjoin RestMethod set}}
	 {self startReader}
      end
      
      meth set(readStopC:     StopChar <= &\n
	       writeAppendVS: AppendVS <= [StopChar])
	 self.ReadStopC     = StopChar
	 self.WriteAppendVS = AppendVS
	 self.SplitP = case StopChar
		       of 0 then Split0
		       [] 128 then Split128
		       else proc {$ S ?S1 ?S2} {Split S StopChar ?S1 ?S2} end
                       end
	 local T in
	    list <- T
	    tail <- T
	 end
      end
      
      meth startReader
         thread
            self.ReaderThread={Thread.this}
            proc {ReadLoop}
               L T in
               Pipe0,read(list:@tail tail:T len:?L)
               if L==0
               then {Thread.terminate {Thread.this}}
               else tail<-T {ReadLoop}
               end
            end
         in
            try
               {Delay 100} % I don't know why, but this seems to be neccessary
	       {ReadLoop}
	    catch system(kernel(terminate...)...) then skip
            [] Ex /*system(os(os "read" 131 ...)...)*/ then
               {Error.printException Ex}
	    finally
               @tail=nil
               self.Terminated=unit
            end
         end
      end

      meth putS(S)
         Pipe0,write(vs:S#self.WriteAppendVS)
      end
      meth getS(?S)
         lock list<-{self.SplitP @list ?S $} end
      end
      meth getBS($)
         {ByteString.make Pipe1,getS($)}
      end
      meth getInt($)
         {StringToInt Pipe1,getS($)}
      end
      meth close
         try
            if {IsFree self.ReaderThread} then
               try self.ReaderThread=unit
               catch failure(...) then skip end
            end
            if self.ReaderThread\=unit then
               {Thread.terminate self.ReaderThread}
               {Wait self.Terminated}
            end
         catch error(kernel(deadThread...)...) then skip 
         finally
            try
               Pipe0, close(kill:true)
            catch Ex then
               {Error.printException Ex}
            end
         end
      end
   end

   CloseMsg = "Process terminated."
   
   class XPipe
      from Pipe1
      
      meth listen(showInfo: ShowInfo <= proc {$ _} skip end
                  onExit:   ExitProc <= proc {$} {ShowInfo CloseMsg} end)
         proc {Loop}
            case {self getS($)} /*Pipe1,getS($)*/ of false then skip
            [] Line then {ShowInfo Line} {Loop} end
         end
      in
         %%{OS.readSelect self.readDesc}
         thread
            try
               {Loop}
            catch connectionClosed then {Browse 'connectionClosed'} skip
            [] Exc then
               {Error.printException Exc}
            finally
               {Browse 'finally'}
               try {ExitProc}
               catch E then {Error.printException E} end
            end
         end
      end
   end

   EndOfMsgChar = 128

   proc {RunClient Cmd Args RemoteHost RemotePort ?KillProc0 ?Done}
      %% create socket and connect
      S = {New XSocket init(readStopC: EndOfMsgChar
                            writeAppendVS: [EndOfMsgChar])}
      {S connect(host:RemoteHost port:RemotePort)}
      %% create pipe
      P = {New XPipe init(cmd:Cmd args:Args
                          readStopC: EndOfMsgChar
                          writeAppendVS: [EndOfMsgChar])}
      %% pipe --> socket forwarder thread
      proc {ShowInfo VS}
         {S putS(VS)}
      end
      Exit
      proc {OnExit}
         {Browse 'Exit'}
         Exit=unit
      end
      {P listen(showInfo:ShowInfo onExit:OnExit)}
      %% socket --> pipe forwarder thread
      T
      thread
         proc {Loop}
            L in {Browse getS(L)}
            case {S getS($)} of false then skip
            [] Line then L=Line {P putS(Line)} {Loop} end
         end
      in
         try
            {Thread.this T}
            {Loop}
         catch system(kernel(terminate...)...) then skip
         [] Exc then
            {Error.printException Exc}
         finally
            Exit=unit
         end
      end
   in
      proc {KillProc0}
         Exit=unit
      end
      thread
         %% wait for exit
         {Browse 'Wait_Exit'}
         {Wait Exit}
         thread % separate thread because this may hang. necessary or not?
            {Browse shutDown(how:[receive])}
            try
               {S shutDown(how:[receive])}
            catch Exc then
               {Error.printException Exc}
            end
         end
         {Delay 100}
         {Browse 'Thread.terminate'(T)}
         try
            {Thread.terminate T}
         catch Exc then
            {Error.printException Exc}
         end
         {Delay 100}
         {Browse 'XPipe,close'}
         try 
            {P close}
         catch Exc then
            {Error.printException Exc}
         end
         {Delay 100}
         {Browse 'Socket,close'}
         try
            {S close}
         catch Exc then
            {Error.printException Exc}
         end
         Done=unit
      end
   end
   
end
