%%
%% process handling (SigKill, SigInt, ...)
%%

functor

import
   Compiler Property OS Type System(printInfo showInfo)

   File 
   at 'x-ozlib://mathweb/share/File.ozf'
   
export
   sig:      Sig
   sigInt:   SigInt
   sigKill:  SigKill
   
   sigF:     DoSig
   sigIntF:  DoSigInt
   sigKillF: DoSigKill

   getTree:  GetTree
   killTree: KillTree
   
define

   %%
   %% get a list of all dependent processes
   %%
   
   PSLinux = "\\ps axjwwwww"    #"|awk '{printf $1 \"#\" $2 \"\\n\"}'|tail +2"
   PSSunOS = "\\ps -Afoppid,pid"#"|awk '{printf $1 \"#\" $2 \"\\n\"}'|tail +2"
   
   proc {ProcessSnapshot ?PairList}
      Tmp = {OS.tmpnam}
      Cmd = case {Property.get 'platform.os'}
            of linux   then PSLinux
            [] solaris then PSSunOS
            end
   in 
      {OS.system Cmd#" >"#Tmp _}
      PairList = {Compiler.virtualStringToValue "["#{File.loadVS Tmp}#"]"}
      try {OS.unlink Tmp}
      catch _ then skip end
   end
   
   /*
   proc {ParentsDict PairList ?D}
      D = {NewDictionary}
      {ForAll PairList
       proc {$ PPID#PID}
          {Dictionary.put D PID PPID}
       end}
   end
   */  
   proc {ChildrenDict PairList ?D}
      D = {NewDictionary}
      {ForAll PairList
       proc {$ PPID#PID}
          {Dictionary.put D PPID PID|{Dictionary.condGet D PPID nil}}
       end}
   end
   
   local
      fun {Enumerate D Ps Tail}
         case Ps of nil then Tail
         [] P|Pr then
            {DoCollect D P {Enumerate D Pr Tail}}
         end
      end
      fun {DoCollect D P Tail}
         P|{Enumerate D {Dictionary.condGet D P nil} Tail} 
      end
   in
      fun {CollectChildren D P}
         {DoCollect D P nil}
      end
   end
   fun {GetTree P}
      {Reverse {CollectChildren {ChildrenDict {ProcessSnapshot}} P}}
   end

   %%
   %% send signals
   %%
   
   %% functions returning "kill" exit code
   local
      fun {CallKill Sig PsVS}
         {OS.system "\\kill -"#Sig#" "#PsVS}
      end
      fun {DoEnumerate Ps}
         case Ps of nil then ''
         [] P|Pr then ' '#P#{DoEnumerate Pr} end
      end
   in
      fun {DoSig Sig Ps}
         case Ps of nil then 0
         [] P|Pr then {CallKill Sig P#{DoEnumerate Pr}}
         else
            {Type.ask.virtualString Ps}
            {CallKill Sig Ps}
         end
      end
   end
   fun {DoSigKill Ps} {DoSig 9     Ps} end
   fun {DoSigInt  Ps} {DoSig 'INT' Ps} end
   
   %% procedures ignoring exit code
   proc {Sig S Ps}   {DoSig S      Ps _} end
   proc {SigKill Ps} {DoSig 9      Ps _} end
   proc {SigInt  Ps} {DoSig 'INT'  Ps _} end
   
   proc {KillTree P}
      PIDs = {GetTree P} in
      {System.printInfo "*** Killing processes:"}
      {ForAll PIDs proc {$ PID} {System.printInfo " "#PID} end}
      {System.showInfo ""}
      {SigKill PIDs}
   end
end












