%%% -*- indent-tabs-mode: nil -*-

functor
   
export
   new : NewDict

   %% wrappers (for compatibility with Dictionary)
   put      : Put
   get      : Get
   condGet  : CondGet
   remove   : Remove
   removeAll: RemoveAll
   member   : IsMember
   keys     : GetKeys
   items    : GetItems
   entries  : GetEntries
   toRecord : ToRecord     %% this should not be used :-)
   clone    : Clone
   isEmpty  : IsEmpty

prepare
   proc {Put        D Key Val}         {D.put Key Val}             end
   fun  {Get        D Key}             {D.get Key}                 end
   fun  {CondGet    D Key DefaultVal}  {D.condGet Key DefaultVal}  end
   proc {Remove     D Key}             {D.remove Key}              end
   proc {RemoveAll  D}                 {D.removeAll}               end
   fun  {IsMember   D Key}             {D.member Key}              end
   fun  {GetKeys    D}                 {D.keys}                    end
   fun  {GetItems   D}                 {D.items}                   end
   fun  {GetEntries D}                 {D.entries}                 end
   fun  {ToRecord   D Label}           {D.toRecord Label}          end
   fun  {Clone      D}                 {D.clone}                   end
   fun  {IsEmpty    D}                 {D.isEmpty}                 end

   VS2A  = VirtualString.toAtom
   VS2BS = ByteString.make
   BS2S  = ByteString.toString

   fun {MakeBS VS}
      if {IsByteString VS} then VS
      else {VS2BS VS} end
   end

   %%
   %% a simple hash function for ByteStrings
   %% (todo: find a better one, implement in C)
   %%
   local
      fun {Hash L N}
         case L
         of H|T then {Hash T (N*16+H) mod 100000000}
         [] nil then N end
      end
   in
      fun {HashBS BS}
         {Hash {BS2S BS} 0}
      end
   end

   %%
   %% simple storage (list of pairs: Key#Value)
   %%
   NotFound = {NewName}
   %%
   fun {Lookup Key L}
      case L of nil then NotFound
      [] (K#V)|T then
         if K==Key then V
         else {Lookup Key T} end
      end
   end
   fun {AddReplace Key Val L}
      case L of nil then (Key#Val)|nil
      [] (P=(K#_))|T then
         if K==Key then (K#Val)|T
         else P|{AddReplace Key Val L} end
      end
   end
   fun {RemoveEntry Key L}
      case L of nil then nil
      [] (P=(K#_))|T then
         if K==Key then T
         else P|{RemoveEntry Key T} end
      end
   end
   
   %%
   %% a dictionary with virtual strings as keys
   %%
   local
      fun {NewDict1 D}
         %%
         L = {NewLock}
         %%
         proc {Put KeyVS Val}
            B={MakeBS KeyVS}
            N={HashBS B}
         in
            lock L then
               D.N := {AddReplace B Val
                       {CondSelect D N nil}}
            end
         end
         %%
         fun {CondGet KeyVS DefaultVal}
            B={MakeBS KeyVS}
            N={HashBS B}
         in
            lock L then
               Val = {Lookup B {CondSelect D N nil}} in
               if {IsDet Val} andthen Val==NotFound then DefaultVal
               else Val end
            end
         end
         %%
         fun {Get KeyVS}
            B={MakeBS KeyVS}
            N={HashBS B}
         in
            lock L then
               Val = {Lookup B {CondSelect D N nil}} in
               if {IsDet Val} andthen Val==NotFound then
                  raise keyNotFound({VS2BS KeyVS}) end
               else Val end
            end
         end
         %%
         proc {Remove KeyVS}
            B={MakeBS KeyVS}
            N={HashBS B}
         in
            lock L then
               case {RemoveEntry B {CondSelect D N nil}}
               of nil then
                  {Dictionary.remove D N}
               [] NewList then
                  D.N := NewList
               end
            end
         end
         %%
         proc {RemoveAll}
            lock L then
               {Dictionary.removeAll D}
            end
         end
         %%
         fun {IsMember KeyVS}
            B={MakeBS KeyVS}
            N={HashBS B}
         in
            lock L then
               Val = {Lookup B {CondSelect D N nil}} in
               {IsDet Val}==false orelse Val\=NotFound
            end
         end
         %%
         fun {GetEntries}
            lock L then
               {FoldR {Dictionary.items D} Append nil}
            end
         end
         %%
         fun {ToRecord Label}
            lock L then
               {List.toRecord Label
                {Map {GetEntries}
                 fun {$ K#V} {VS2A K}#V end}}
            end
         end
         %%
         fun {Clone}
            lock L then
               D1 = {Dictionary.clone D} in
               {NewDict1 D1}
            end
         end
      in
         dict(
            put      : Put
            get      : Get
            condGet  : CondGet
            remove   : Remove
            removeAll: RemoveAll
            member   : IsMember
            entries  : GetEntries
            keys     : fun {$} {Map {GetEntries} fun {$ K#_} K end} end
            items    : fun {$} {Map {GetEntries} fun {$ _#V} V end} end
            toRecord : ToRecord
            clone    : Clone
            isEmpty  : fun {$} {Dictionary.isEmpty D} end
            %%dictionary: D
            )
      end
   in
      fun {NewDict}
         D = {NewDictionary} in
         {NewDict1 D}
      end
   end
end




