%%
%% Text: a sequence of text lines
%%

functor
   
export          
   make:            Make             %% VirtualString --> Text
   append:          Append           %% Text1 x Text2 --> Text3
   toString:        ToString         %% Text --> String
   toVirtualString: ToVirtualString  %% Text --> VirtualString
   toByteString:    ToByteString     %% Text --> ByteString

   shift:           Shift            %% fun {Shift  Text LinePrefix}
   shiftN:          ShiftN           %% fun {ShiftN Text N}

   spacelines:      Spacelines       %% proc {Spacelines N ?Text}
   addSpacelines:   AddSpacelines    %% fun {AddSpacelines Text N}
   addSpaceline:    AddSpaceline     %% fun {AddSpaceline Text}

   splitFirstLine:  SplitFirstLine   %% proc {SplitFirstLine Text ?Line ?Rest}
   forAll:          ForAll           %% proc {ForAll Text P}
   
   buffer: Buffer
   %%         meth init(Text <= nil)
   %%         meth initVS(VS)
   %%         meth get($)   
   %%         meth append(Text)
   %%         meth appendLine(Line)
   %%         meth toVirtualString($)
   %%         meth toString($)   
   %%         meth toByteString($)
   %%         meth shift(LinePrefix)
   %%         meth shiftN(N)
   %%         meth addSpaceline
   %%         meth addSpacelines(N)
   %%         meth splitFirstLine(?FirstLine)
   %%         meth getS($)
   %%         meth forAll(P)
   %%         meth clear
   
define

   %%
   %% A text is currently implemented as a list of strings.
   %%
   
   local
      proc {SplitFirstLine Xs ?S ?Rest}
         case Xs
         of nil      then S=nil Rest=nil
         [] &\n|Xr   then S=nil Rest=Xr
         elseof X|Xr then S=X|{SplitFirstLine Xr $ ?Rest}
         end
      end
      fun {SplitAllLines S}
         if S == nil then nil
         else
            FirstLine RestLines in
            {SplitAllLines {SplitFirstLine S ?FirstLine $} ?RestLines}
            FirstLine|RestLines   
         end
      end
   in
      %%
      %% If VS does not end with <newline>, then it is appended.
      %%
      
      fun {Make VS}
         {SplitAllLines {VirtualString.toString VS}}
      end
   end
   
   fun {ToVirtualString Text}
      {FoldL Text fun {$ VS Line} VS#Line#"\n" end nil}
   end
   fun {ToString Text}
      {VirtualString.toString {ToVirtualString Text}}
   end
   fun {ToByteString Text}
      {VirtualString.toByteString {ToVirtualString Text}}
   end
   
   fun {Shift Text LinePrefix}
      {Map Text fun {$ Line} LinePrefix#Line end}
   end
   fun {ShiftN Text N}
      NSpaces = {List.make N}
      {ForAll NSpaces proc {$ C} C=&  end}
   in
      {Map Text fun {$ Line} NSpaces#Line end}
   end
   fun {Append Text1 Text2}
      {List.append Text1 Text2}
   end
   proc {Spacelines N ?Text}
      {List.make N ?Text}
      {ForAll Text proc {$ Line} Line="" end}
   end
   fun {AddSpacelines Text N}
      {Append Text {Spacelines N}}
   end
   fun {AddSpaceline Text}
      {AddSpacelines Text 1}
   end
   proc {SplitFirstLine Text ?Line ?Rest}
      case Text
      of nil then Line=false Rest=nil
      [] L1|Lr then Line=L1 Rest=Lr
      end
   end
   proc {ForAll Text P}
      {List.forAll Text P}
   end
   
   class Buffer
      attr
         text: nil
         
      meth init(Text <= nil)
         text <- Text
      end
      meth initVS(VS)
         text <- {Make VS}
      end
      meth get($)
         @text
      end
      meth append(Text)
         text <- {Append @text Text}
      end
      meth appendLine(Line)
	 text <- {Append @text [Line]}
      end
      meth toVirtualString($)
         {ToVirtualString @text}
      end
      meth toString($)
         {ToString @text}
      end
      meth toByteString($)
         {ToByteString @text}
      end
      meth shift(LinePrefix)
         text <- {Shift @text LinePrefix}
      end
      meth shiftN(N)
         text <- {ShiftN @text N}
      end
      meth addSpaceline
         text <- {AddSpaceline @text}
      end
      meth addSpacelines(N)
         text <- {AddSpacelines @text N}
      end
      meth splitFirstLine(?FirstLine)
         text <- {SplitFirstLine @text ?FirstLine $}
      end
      meth getS($)
         Buffer, splitFirstLine($)
      end
      meth forAll(P)
         {ForAll @text P}
      end
      meth clear
	 text <- nil
      end
         
   end
   
end








