%%% -*- indent-tabs-mode: nil -*-
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% XApplication.oz ; This file is part of the MathWeb system
%
%   major updates: 01.02.2001
%
%   Authors: Andreas Franke
%
%   email: afranke@ags.uni-sb.de
%   
% For information about this program, write to:                          
%   MathWeb-Project                                                        
%   AG Siekmann/FB nformatik                                            
%   Universitaet des Saarlandes                                          
%   Bau 36, 4. Stock                                                     
%   D-66041 Saarbruecken                                                 
%   Germany    
%
% For information about the newest version of MathWeb, see 
%   http://www.mathweb.org/mathweb
%
% See the file "LICENSE" or
%    http://www.mathweb.org/mathweb/LICENSE.html
% for information on usage and redistribution
% of this file, and for a DISCLAIMER OF ALL
% WARRANTIES.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%
%%% thread registry and application emulation for mozart-shell 'applications'
%%%

functor

import
   Application Error Module Pickle Property Open OS System

export
   parseCmdArgs: ParseCmdArgs
   parseCgiArgs: ParseCgiArgs
   
   counter: Counter
   child:   Child
   output:  Output
   debug:   Debug
   threads: Threads
   args:    Args
   exit:    Exit
   manager: ApplicationManager

   wrapper: Wrapper
   
define
   
   %% counter
   
   class Counter
      attr
         counter:0
         
      meth init(N0<=0)
         counter <- N0
      end         
      meth next(?N)
         N = @counter + 1
         counter <- N
      end
   end
   
   %%
   %% child
   %%

   local skip in

      class Child
         
         feat
            toplevel
            parent
            
         meth initChild(parent: Parent <= false)
            self.toplevel = Parent == false
            self.parent   = if self.toplevel then self else Parent end
         end
         meth toplevel($)
            self.toplevel
         end
         meth parent($)
            self.parent
         end
         meth otherwise(M)
            if self.toplevel
            then raise error(object(lookup 'Child' {Label M})) end
            else {self.parent M}
            end
         end
      end

   end

   %%
   %% ID management
   %%

   local skip in
      
      class ID from Child

         feat
            id
            
         meth initID(ThrID <= 0)
            self.id = ThrID
         end
         meth thisStepID($)
            self.id
         end
         meth recursiveID($)
            if Child,toplevel($)
            then ""
            else {Child,parent($) recursiveID($)}#'.'#self.id
            end
         end
         meth thisID($)
            case {VirtualString.toString ID,recursiveID($)}
            of "" then 0
            [] &.|Rest then {StringToAtom Rest}
            end
         end
         meth expandID(RelID $)
            &.|Rest = {VirtualString.toString (ID,recursiveID($))#'.'#RelID}
         in
            {StringToAtom Rest}
         end
      end
   end

   %%
   %% application exit
   %%

   local skip in
      
      class Exit

         feat
            exit
         attr
            exit: Application.exit
            
         meth initExit(ExitProc <= @exit)
            exit <- ExitProc
            self.exit = proc {$ Status} Exit, exit(Status) end
         end
         
         meth exit(Status)
            {@exit Status}
         end         
      end
   end
   
   %%
   %% thread registry
   %%
   
   local
      proc {THREAD_NOT_FOUND ThrID}
         raise threadRegistry(threadNotFound threadID:ThrID) end
      end
     
      PropertyLock = {NewLock}
      proc {PropertyPut ChildInfo}
         lock PropertyLock then
            {Property.put 'x-application.children'
             ChildInfo | {Property.condGet 'x-application.children' nil}}    
         end
      end
      proc {PropertyRemove ChildInfo}
         lock PropertyLock then
            {Property.put 'x-application.children'
             {List.subtract {Property.get 'x-application.children'} ChildInfo}}
         end
      end
      
      proc {Skip1 _} skip end

      fun {NotUnit X} {Not ({IsDet X} andthen {IsUnit X})} end

      %% thread ID conversion 1.2.3 --> [1 2 3]
      local
         proc {ReadTillPoint S ?Str ?Rest}
            case S of nil then Str=nil Rest=nil
            [] C|Sr then
               if C == &.
               then Str=nil Rest=Sr 
               else Str=C|{ReadTillPoint Sr $ ?Rest}
               end
            end
         end
         fun {SplitString S}
            if S == nil then nil
            else
               FirstS RestS in
               {ReadTillPoint S ?FirstS ?RestS}
               FirstS|{SplitString RestS}
            end
         end
      in
         fun {ThreadID2List VS}
            {Map {SplitString {VirtualString.toString VS}} StringToInt}
         end
      end
            
      %% thread ID conversion [1 2 3] --> 1.2.3
      local
         fun {Enumerate Xs Sep}
            fun {DoEnumerate Xs}
               case Xs of nil then ''
               [] X|Xr then
                  Sep#X#{DoEnumerate Xr}
               end
            end
         in
            case Xs of nil then
               raise enumerate(emptyList) end
            [] X|Xr then
               X#{DoEnumerate Xr}
            end
         end
      in
         fun {List2ThreadID L}
            {VirtualString.toAtom {Enumerate L '.'}}
         end
      end

      NotFound = {NewName}
   in
      
      class Threads

         from
            Child ID Exit
         
         prop
            locking
            
         feat
            counter
            registry
            modMan
            
         attr
            empty: unit
            
         meth initThreads(ThrID <= unit
                          exit:   ExitProc
                          parent: Parent   <= unit
                          modMan: ModMan   <= {New Module.manager init})
            
            Child, {Record.filter initChild(parent:Parent) NotUnit}
            ID,    {Record.filter initID(ThrID)            NotUnit}
            Exit,                 initExit(ExitProc)
            self.counter  = {New Counter init(0)}
            self.registry = {NewDictionary}
            self.modMan   = ModMan
            empty <- unit
         end
         
         %% dictionary operations
         meth keys($)
            lock
               /*
               {ForAll {Dictionary.entries self.registry}
                proc {$ ThrID#Entry}
                   if {Thread.state Entry.'thread'}==terminated
                   then {Dictionary.remove self.registry ThrID}
                   end
                end}
               */
               {Sort {Dictionary.keys self.registry} Value.'=<'}
            end
         end
         meth items($)
            lock {Dictionary.items self.registry} end
         end
         meth entries($)
            lock {Dictionary.entries self.registry} end
         end
         meth toRecord($)
            lock {Dictionary.toRecord threads self.registry} end
         end
         %% condGet
         meth condGetInt(ThrID Default $)
            lock
               case {Dictionary.condGet self.registry ThrID false}
               of false then Default
               elseof Entry then Entry
               end
            end
         end
         meth condGetList(ThrID Default $)
            lock
               ID1|Rest = ThrID
            in
               case {Dictionary.condGet self.registry ID1 false}
               of false then Default
               elseof Entry then
                  if Rest==nil then
                     Entry
                  elseif {HasFeature Entry child} then
                     {Entry.child condGetList(Rest Default $)}
                  else
                     Default
                  end
               end
            end
         end
         meth condGetAtom(ThrID Default $)
            Threads, condGetList({ThreadID2List ThrID} Default $)
         end
         meth condGet(ThrID Default $)
            lock
               if {IsInt ThrID} then
                  Threads, condGetInt(ThrID Default $)
               elseif {IsAtom ThrID} then
                  Threads, condGetAtom(ThrID Default $)
               else
                  Threads, condGetList(ThrID Default $)
               end
            end
         end
         %% member
         meth memberInt(ThrID $)
            lock {Dictionary.member self.registry ThrID} end
         end
         meth memberList(ThrID $)
            Threads, condGetList(ThrID false $) \= false
         end
         meth memberAtom(ThrID $)
            Threads, memberList({ThreadID2List ThrID} $)
         end
         meth member(ThrID $)
            lock
               if {IsInt ThrID} then
                  Threads, memberInt(ThrID $)
               elseif {IsAtom ThrID} then
                  Threads, memberAtom(ThrID $)
               else
                  Threads, memberList(ThrID $)
               end
            end
         end
         %% CondGetGeneric
         meth CondGetGenericInt(ThrID F Default $ default:Def<=Default)
            lock
               case {Dictionary.condGet self.registry ThrID false}
               of false then Default
               elseof Entry then
                  {Value.condSelect Entry F Def}
               end
            end
         end
         meth CondGetGenericList(ThrID F Default $ default:Def<=Default)
            case Threads, condGetList(ThrID false $)
            of false then Default
            elseof Entry then
               {Value.condSelect Entry F Def}
            end
         end
         meth CondGetGeneric(ThrID F Default $ default:Def<=Default)
            lock
               if {IsInt ThrID} then
                  Threads, CondGetGenericInt(ThrID F Default $ default:Def)
               elseif {IsAtom ThrID} then
                  ThrID1 = {ThreadID2List ThrID} in
                  Threads, CondGetGenericList(ThrID1 F Default $ default:Def)
               else
                  Threads, CondGetGenericList(ThrID F Default $ default:Def)
               end
            end
         end
         %% condGetThread, condGetInfo, condGetValue, condGetChild
         meth condGetThread(ThrID Default $)
            Threads, CondGetGeneric(ThrID 'thread' Default $)
         end
         meth condGetInfo(ThrID Default $ default:Def<=Default)
            Threads, CondGetGeneric(ThrID info Default $ default:Def)
         end
         meth condGetValue(ThrID Default $ default:Def<=Default)
            Threads, CondGetGeneric(ThrID value Default $ default:Def)
         end
         meth condGetChild(ThrID Default $ default:Def<=Default)
            Threads, CondGetGeneric(ThrID child Default $ default:Def)
         end
         
         %% get
         meth getInt(ThrID $)
            lock
               case {Dictionary.condGet self.registry ThrID false}
               of false then {THREAD_NOT_FOUND ThrID} false
               elseof Entry then
                  Entry
               end
            end
         end
         meth getList(ThrID $)
            case Threads, condGetList(ThrID false $)
            of false then {THREAD_NOT_FOUND {List2ThreadID ThrID}} false
            elseof Entry then
               Entry
            end
         end
         meth getAtom(ThrID $)
            Threads, getList({ThreadID2List ThrID} $)
         end
         meth get(ThrID $)
            lock
               if {IsInt ThrID} then
                  Threads, getInt(ThrID $)
               elseif {IsAtom ThrID} then
                  Threads, getAtom(ThrID $)
               else
                  Threads, getList(ThrID $)
               end
            end
         end
         %% GetGeneric
         meth GetGenericInt(ThrID F $ default:Default<=false)
            lock
               case {Dictionary.condGet self.registry ThrID NotFound}
               of !NotFound then
                  {THREAD_NOT_FOUND ThrID} dummy
               elseof Entry then
                  {Value.condSelect Entry F Default}
               end
            end
         end
         meth GetGenericList(ThrID F $ default:Default<=false)
            case Threads, condGetList(ThrID NotFound $) of !NotFound then
               {THREAD_NOT_FOUND {List2ThreadID ThrID}} dummy
            elseof Entry then
               {Value.condSelect Entry F Default}
            end
         end
         meth GetGeneric(ThrID F $ default:Default<=false)
            lock
               if {IsInt ThrID} then
                  Threads, GetGenericInt(ThrID F $ default:Default)
               elseif {IsAtom ThrID} then
                  ThrID1 = {ThreadID2List ThrID} in
                  Threads, GetGenericList(ThrID1 F $ default:Default)
               else
                  Threads, GetGenericList(ThrID F $ default:Default)
               end
            end
         end
         %% getThread, getStart, getInfo, getValue, getChild
         meth getThread(ThrID $)
            Threads, GetGeneric(ThrID 'thread' $)
         end
         meth getStart(ThrID $)
            Threads, GetGeneric(ThrID start $)
         end
         meth getInfo(ThrID $ default:Default<=unit)
            Threads, GetGeneric(ThrID info $ default:Default)
         end
         meth getValue(ThrID $ default:Default<=unit)
            Threads, GetGeneric(ThrID value $ default:Default)
         end
         meth getChild(ThrID $ default:Default<=false)
            Threads, GetGeneric(ThrID child $ default:Default)
         end

         %% Put & Remove
         meth Put(ThrID Thr Entry<=dummy)
            lock
               {Dictionary.put self.registry ThrID
                {Adjoin Entry entry('thread':Thr)}}
            end
         end
         meth Remove(ThrID)
            lock
               {Dictionary.remove self.registry ThrID}
               if Threads, isEmpty($) then
                  @empty=unit
               end               
            end
         end
         %% forAll...
         meth forAll(Proc)
            lock {ForAll {Dictionary.keys self.registry} Proc} end
         end
         meth forAllThreads(Proc)
            lock
               {ForAll {Dictionary.items self.registry}
                proc {$ Entry} {Proc Entry.'thread'} end}
            end
         end
         meth forAllInd(Proc)
            lock
               {ForAll {Dictionary.entries self.registry}
                proc {$ ThrID#Entry} {Proc ThrID Entry} end}
            end
         end
         meth forAllIndThreads(Proc)
            lock
               {ForAll {Dictionary.entries self.registry}
                proc {$ ThrID#Entry} {Proc ThrID Entry.'thread'} end}
            end
         end
         %% isEmpty
         meth isEmpty($)
            Threads, keys($) == nil
         end
         
         %% register, unregister, emptyFlag
         meth Register(?ThrID Entry<=dummy)
            {self printDebug('Registering thread...')}
            lock
               ThrID = {self.counter next($)}
               Threads, Put(ThrID {Thread.this}
                            {AdjoinAt Entry start {OS.localTime}})
               if {IsFree @empty}==false then
                  empty <- _
               end
            end
            {self showDebug(' done. ID = '#ThrID)}
         end
         meth RegisterAdjoin(ThrID ValRec)
            {self showDebug('Registering optional features (ID = '#ThrID#'):')}
            {Record.forAllInd ValRec
             proc {$ F V} {self printDebug(F#": ")} {self showD(V)} end}
            lock
               OldEntry = {Dictionary.get self.registry ThrID}
               NewEntry = {Adjoin ValRec OldEntry}
            in
               {Dictionary.put self.registry ThrID NewEntry}
            end
            {self showDebug('Registering optional features... done.')}
         end
         meth RegisterChild(ThrID Child)
            {self RegisterAdjoin(ThrID unit(child:Child))}
         end
         meth Unregister(ThrID)
            {self printDebug('Unregistering thread '#ThrID#'...')}
            try
               lock
                  Entry = Threads, getInt(ThrID $)
               in
                  if {HasFeature Entry child} then
                     %% {Entry.child tryTerminateAll}
                     thread
                        {Wait {Entry.child empty($)}}
                        Threads, Remove(ThrID)
                     end
                  else
                     Threads, Remove(ThrID)
                  end
               end
            catch Err then
               {self showDebug('*** ERROR'#
                               'while unregistering thread '#ThrID#': ')}
               {self showD(Err)}
               {self showDebug('retrying in 10 seconds...')}
               try {Delay 10000}
               catch _ then skip
               finally Threads, Unregister(ThrID) end
            end
            {self showDebug(' done.')}
         end
         meth empty($)
            lock @empty end
         end

         %% thread operations

         meth state(ThrID $)
            lock {Thread.state Threads,getThread(ThrID $)} end
         end
         meth getPriority(ThrID $)
            lock {Thread.getPriority Threads,getThread(ThrID $)} end
         end
         meth setPriority(ThrID Priority)
            lock {Thread.setPriority Threads,getThread(ThrID $)
                  Priority} end
         end
         meth suspend(ThrID)
            lock {Thread.suspend Threads,getThread(ThrID $)} end
         end
         meth isSuspended(ThrID $)
            lock {Thread.isSuspended Threads,getThread(ThrID $)} end
         end
         meth resume(ThrID)
            lock {Thread.resume Threads,getThread(ThrID $)} end
         end
         meth injectException(ThrID Exception)
            lock {Thread.injectException Threads,getThread(ThrID $)
                  Exception} end
         end
         
         %% safe thread operations

         meth tryState(ThrID $)
            lock
               try Threads, state(ThrID $)
               catch _ then {Delay 1} unit end
            end
         end
         meth tryGetPriority(ThrID $)
            lock
               try Threads, getPriority(ThrID $)
               catch _ then {Delay 1} unit end
            end
         end
         meth trySetPriority(ThrID Priority)
            lock
               try Threads, setPriority(ThrID Priority)
               catch _ then {Delay 1} skip end
            end
         end
         meth trySuspend(ThrID)
            lock
               try Threads, suspend(ThrID)
               catch _ then {Delay 1} skip end
            end
         end
         meth tryIsSuspended(ThrID $)
            lock
               try Threads, isSuspended(ThrID $)
               catch _ then {Delay 1} unit end
            end
         end
         meth tryResume(ThrID)
            lock
               try Threads, resume(ThrID)
               catch _ then {Delay 1} skip end
            end
         end
         meth tryInjectException(ThrID E)
            lock
               try Threads, injectException(ThrID E)
               catch _ then {Delay 1} skip end
            end
         end
         
         %% creating new threads

         meth defaultHandler(ThrID Exc info:Info<=unit)
            `%5`="%%%%%" `%10`=`%5`#`%5` `%20`=`%10`#`%10` in
            lock {self lockError($)} then
               {self showError('')}
               {self showError(`%20`#`%20`#`%20`#`%5`)}
               {self printError("%%% Error in thread ")} {self showE(ThrID)}
               if Info\=unit then {self showError("%%% "#Info)} end
               {self showError(`%20`#`%20`#`%20`#`%5`)}
               {self printExc(Exc)}
            end
         end
         
         meth newThread(Proc ?ThrID <= _
                        detach:   Detach  <= false
                        info:     Info    <= unit
                        value:    Val     <= unit
                        exit:     ?Exit   <= _
                        onError:  DoError <= _ 
                        onExit:   DoExit  <= Skip1) = M
            
            if {IsFree DoError} then
               proc {DoError Exc}
                  {self defaultHandler(ThrID Exc info:Info)}
               end
            end
            
            if Detach == true andthen self.toplevel == false then
               {self.parent {Record.subtract M detach}}
            else
               Registered in
               thread
                  try
                     Entry = {Record.filter entry(info:  Info
                                                  value: Val) NotUnit}
                  in
                     Threads, Register(?ThrID Entry)
                  catch Err then
                     Exit = Err
                  finally
                     Registered = unit
                  end

                  if {IsDet Exit}==false then
                     ExitStatus in
                     try
                        case {Procedure.arity Proc} 
                        of 0 then
                           {Proc}
                        [] 1 then
                           proc {ExitProc ExitStatus}
                              thread Threads,tryKill(ThrID exit:ExitStatus) end
                              {Thread.suspend {Thread.this}}
                           end
                           Chld = {New Threads
                                   initThreads(ThrID
                                               parent: self
                                               exit:   ExitProc
                                               modMan: self.modMan)}
                           Threads, RegisterChild(ThrID Chld)
                        in
                           {Proc Chld}
                        end
                        ExitStatus = unit
                     catch Err then
                        case Err of system(kernel(terminate exit:X...)...) then
                           ExitStatus = X
                        elseof system(kernel(terminate...)...) then
                           ExitStatus = unit
                        else
                           ExitStatus = Err
                           try {DoError Err}
                           catch _ then skip end
                        end
                     finally
                        Threads, Unregister(ThrID)
                        try Exit = ExitStatus
                        catch _ then skip end
                        try {DoExit ExitStatus}
                        catch _ then skip end
                     end
                  end
               end
               {Wait Registered}
            end
         end
   
         meth startApplication(URL ?ThrID <= _
                               'functor': Functor  <= {Pickle.load URL}
                               args:      Args     <= nil
                               modMan:    ModMan   <= self.modMan
                               detach:    Detach   <= false
                               info:      Info     <= unit
                               exit:      ?Exit    <= _
                               onError:   DoError  <= _
                               onExit:    DoExit   <= Skip1) = M
            
            if {IsFree DoError} then
               proc {DoError Exc}
                  {self defaultHandler(ThrID Exc info:Info)}
               end
            end
           
            if Detach == true andthen self.toplevel == false then
               {self.parent {Record.subtract M detach}}
            else
               if {HasFeature Functor.'import' 'Application'} then
                  raise error(ap('`Application` imported' url:URL)) end
               end
               Registered in
               thread
                  try
                     Entry = {Record.filter entry(info:Info) NotUnit}
                  in
                     Threads, Register(?ThrID Entry)
                  catch Err then
                     Exit = Err
                  finally
                     Registered = unit
                  end
                  
                  if {IsDet Exit}==false then
                     ExitStatus ChildInfo
                     proc {DoFirst Child}
                        Threads, RegisterChild(ThrID Child)
                     end
                  in
                     try
                        ChildInfo = app('thread':  {Thread.this}
                                        'functor': Functor
                                        doFirst:   DoFirst
                                        parent:    self
                                        threadID:  ThrID
                                        url:       URL
                                        args:      Args
                                        modMan:    ModMan
                                        info:      Info)
                        {PropertyPut ChildInfo}
                        {ModMan apply(url:URL Functor)}
                        ExitStatus = unit
                     catch Err then
                        case Err of system(kernel(terminate exit:X...)...) then
                           ExitStatus = X
                        elseof system(kernel(terminate...)...) then
                           ExitStatus = unit
                        else
                           ExitStatus = Err
                           try {DoError Err}
                           catch _ then skip end
                        end
                     finally
                        Threads, Unregister(ThrID)
                        try Exit = ExitStatus
                        catch _ then skip end
                        try {DoExit ExitStatus}
                        catch _ then skip end
                        {PropertyRemove ChildInfo}
                     end
                  end
               end
               {Wait Registered}
            end
         end
         
         meth newServer(C Init ?PS ?ThrID<=_
                        detach:  Detach  <= true
                        info:    Info    <= '<unknown>.server'
                        onError: DoError <= _
                        onExit:  DoExit  <= Skip1)
            
            S CellS P ExtendedClass Obj Handle DoThread 
         in
            if {IsFree DoError} then
               proc {DoError Exc}
                  {self defaultHandler(ThrID Exc info:Info)}
               end
            end
            P  = {Port.new S}
            CellS = {NewCell S} % workaround to keep S garbage collectable
            PS = proc {$ M}
                    {Port.send P M}
                 end
            ExtendedClass = {Class.new [C] a() f(serverProc: PS
                                                 serverPort: P
                                                 appManager: self
                                                 localRegistry: _) nil}
            Obj = {New ExtendedClass Init}
            proc {Handle Meth Exc}
               `%5`="%%%%%" `%10`=`%5`#`%5` `%20`=`%10`#`%10` in
               lock {self lockError($)} then
                  {self showError('')}
                  {self showError(`%20`#`%20`#`%20`#`%5`)}
                  {self printError("%%% Error in server-thread ")}
                  {self showE(ThrID)}
                  if Info\=unit then {self showError("%%% "#Info)} end
                  {self showError("%%% during execution of method: ")}
                  {self printError("%%% ")} {self showE(Meth)}
                  {self showError(`%20`#`%20`#`%20`#`%5`)}
                  {self printExc(Exc)}
               end
            end
            proc {DoThread LocalRegistry}
               S = {Exchange CellS $ nil} % workaround, see above
            in
               Obj.localRegistry = LocalRegistry
               try
                  {ForAll S
                   proc {$ Meth}
                      try
                         {Obj Meth}
                      catch Err then
                         case Err of system(kernel(terminate...)...) then
                            raise Err end
                         else
                            try {Obj 'EXCEPTIONHANDLER'(Err 'in':Meth)}
                            catch _ then
                               {Handle Meth Err}
                            end
                         end
                      end
                   end}
               catch Err = system(kernel(terminate...)...) then
                  try {Obj 'SHUTDOWN'}
                  catch _ then skip end
                  raise Err end
               end
            end
            Threads, newThread(DoThread ?ThrID
                               detach:  Detach
                               info:    Info
                               value:   Obj
                               onError: DoError
                               onExit:  DoExit)
         end
         meth sendObject(ThrID M)
            {Threads,getValue(ThrID $) M}
         end
         meth sendServer(ThrID M)
            {(Threads,getValue(ThrID $)).serverProc M}
         end
         meth sendPort(ThrID M)
            {Port.send (Threads,getValue(ThrID $)).serverPort M}
         end
         
         %% terminate threads
         
         meth terminate(ThrID
                        exit: Exit <= unit    % exit status (unit=regular)
                        info: Info <= unit    % info about situation/reason
                       )
            E in
            if Info == unit
            then E = system(kernel(terminate exit:Exit))   
            else E = system(kernel(terminate exit:Exit info:Info)) 
            end            
            Threads, injectException(ThrID E)
         end
         meth tryTerminate(ThrID exit:Exit<=unit info:Info<=unit)
            try Threads, terminate(ThrID exit:Exit info:Info)
            catch _ then {Delay 1} skip end
         end
         meth terminateAll(exit: Exit <= unit
                           info: Info <= unit)
            {ForAll Threads, keys($)
             proc {$ ThrID}
                Threads, tryTerminate(ThrID exit:Exit info:Info)
             end}
         end
         meth tryTerminateAll(exit: Exit <= unit
                              info: Info <= unit)
            try Threads, terminateAll(exit:Exit info:Info)
            catch _ then {Delay 1} skip end
         end

         %% kill threads
         
         meth kill(ThrID
                   exit:       Exit        <= _      % exit status
                   delay:      DelayTime   <= 100    % time (ms) between tries
                   retry:      MaxRetry    <= 9      % max. number of tries
                   abort:      Abort       <= _      % abort kill loop
                   terminated: ?Terminated <= _      % (bool) success?
                  )
            Entry = Threads, get(ThrID $)
            Thr   = Entry.'thread'
            
            if Thr == {Thread.this} then
               raise threads(suicide threadID:ThrID 'thread':Thr) end
            end
            
            if {HasFeature Entry child} then
               try {Thread.suspend Thr}
               catch _ then skip end               
               {Entry.child killAll(delay: DelayTime
                                    retry: MaxRetry
                                    abort: Abort)}
            end

            fun {Ord N}
               case N
               of 1 then 'st'
               [] 2 then 'nd'
               [] 3 then 'rd'
               else 'th'
               end
            end
            proc {TryKill N}
               ThreadState = {Thread.state Thr}
            in
               if {IsDet Abort} orelse
                  N > MaxRetry orelse
                  ThreadState == terminated
               then
                  Terminated = (ThreadState == terminated)
               else
                  TryKillMsg = 'Trying to kill thread '#ThrID#'.'
                  StateMsg   = 'Thread.state = '#ThreadState#'.'
                  InjectMsg  = if N == 0
                               then 'Injecting exception...'
                               else N#{Ord N}#' retry...'
                               end
               in
                  {self showDebug(TryKillMsg#' '#StateMsg#' '#InjectMsg)}
                  try
                     if {IsDet Exit}
                     then Threads, terminate(ThrID exit:Exit)
                     else Threads, terminate(ThrID exit:killed)
                     end
                     {Delay DelayTime}
                     {TryKill N+1}
                  catch error(kernel(deadThread...)...) then
                     Terminated = true
                  end
               end
            end
         in
            try {Thread.setPriority Thr low}
            catch _ then {Delay 1} skip end
            {TryKill 0}
            if Terminated
            then {self showDebug('Thread '#ThrID#' terminated.')}
            else {self showDebug('Thread '#ThrID#' not terminated.')}
            end           
         end
         meth tryKill(ThrID
                      exit:       Exit        <= _
                      delay:      DelayTime   <= 100
                      retry:      MaxRetry    <= 9
                      abort:      Abort       <= _
                      terminated: ?Terminated <= _)
            try
               Threads, kill(ThrID
                             exit:       Exit
                             delay:      DelayTime
                             retry:      MaxRetry
                             abort:      Abort
                             terminated: ?Terminated)
            catch _ then {Delay 1} skip end
         end
         
         meth killAll(exit:    Exit       <= _
                      delay:   DelayTime  <= 100
                      retry:   MaxRetry   <= 9
                      abort:   Abort      <= _
                      timeOut: TimeOut    <= unit
                      success: ?Success   <= unit)
            AlarmTimeOut
            SuccessAll
         in
            SuccessAll = {FoldL Threads, keys($)
                          proc {$ SuccessOld ThrID ?SuccessNew}
                             thread
                                Terminated in
                                Threads, tryKill(ThrID
                                                 exit:      Exit
                                                 delay:     DelayTime
                                                 retry:     MaxRetry
                                                 abort:     Abort
                                                 terminated:?Terminated)
                                {And SuccessOld Terminated ?SuccessNew}
                             end
                          end
                          true}

            if TimeOut \= unit then
               AlarmTimeOut = {Alarm TimeOut}
            end
            
            {WaitOr SuccessAll AlarmTimeOut}
            if Threads, isEmpty($) == false
            then
               if {IsDet Success} then
                  raise kill(timeOut threads:Threads,keys($)) end
               else Success = SuccessAll
               end
            else
               if {IsDet Success} == false then
                  Success = SuccessAll
               end
            end
         end
         meth tryKillAll(exit:    Exit      <= _
                         delay:   DelayTime <= 100
                         retry:   MaxRetry  <= 9
                         abort:   Abort     <= _
                         timeOut: TimeOut   <= unit
                         success: ?Success  <= unit)
            try
               Threads, killAll(exit:    Exit
                                delay:   DelayTime
                                retry:   MaxRetry
                                abort:   Abort
                                timeOut: TimeOut
                                success: ?Success)
            catch _ then {Delay 1} skip end
         end
      end
   end

   %%
   %% environment
   %%
  
   local
      
      NotFound = {NewName}
      
      fun {GetEnvA Env KeyA}
         case {Dictionary.condGet Env KeyA NotFound} of !NotFound
         then
            case {Dictionary.condGet Env 'INHERITS-FROM' false} of false
            then
               {OS.getEnv KeyA}
            elseof ParentEnv then
               {GetEnvA ParentEnv KeyA}
            end
         elseof Val then Val
         end
      end

      proc {PutEnvA Env KeyA Val}
         {Dictionary.put Env KeyA Val}  
      end

      VS2A = VirtualString.toAtom
      
   in

      %%
      %% local environment
      %%
      
      class LocalEnv
         
         feat
            getLocalEnv
            putLocalEnv
            env
            
         meth initLocalEnv(env:          Env       <= {NewDictionary}
                           inheritsFrom: ParentEnv <= false)

            self.env = Env

            if ParentEnv \= false then
               {Dictionary.put Env 'INHERITS-FROM' ParentEnv}
            end

            self.getLocalEnv = fun {$ KeyVS}
                                  LocalEnv, getLocalEnv(KeyVS $)
                               end
            self.putLocalEnv = proc {$ KeyVS Val}
                                  LocalEnv, putLocalEnv(KeyVS Val)
                               end
         end
         
         meth getLocalEnv(KeyVS $)
            {GetEnvA self.env {VS2A KeyVS}}
         end
         meth putLocalEnv(KeyVS Val)
            {PutEnvA self.env {VS2A KeyVS} Val}
         end
      end

      %%
      %% environment access
      %%

      class EnvAccess

         feat
            getEnv
            putEnv
            
         attr
            getEnv: OS.getEnv
            putEnv: OS.putEnv

         meth initEnvAccess(getEnv: GetEnv <= @getEnv
                            putEnv: PutEnv <= @putEnv)
      
            getEnv <- GetEnv
            putEnv <- PutEnv
            
            self.getEnv = fun  {$ Var}     EnvAccess, getEnv(Var $)   end
            self.putEnv = proc {$ Var Val} EnvAccess, putEnv(Var Val) end
         end
         
         meth getEnv(KeyVS $)
            {@getEnv KeyVS}
         end
         meth putEnv(KeyVS Val)
            {@putEnv KeyVS Val}
         end 
      end
      
   end

   %%
   %% standard output, error output & debug output
   %%
   
   local

      fun {ExceptionToVirtualString Exc}
         {Error.messageToVirtualString {Error.exceptionToMessage Exc}}
      end

      VS2S = VirtualString.toString
      
      V2VS = Value.toVirtualString

      proc {Skip0} skip end

      if {Property.condGet 'writelock' unit} == unit then
         {Property.put 'writelock' {NewLock}}
      end
   in 

      class Output

         feat
            printInfo  showInfo  print  show
            printError showError printE showE
            
         attr
            printInfo:  System.printInfo
            printError: System.printError

            lockInfo:  {Property.get 'writelock'}
            lockError: {Property.get 'writelock'}
               
            closeInfo:  Skip0
            closeError: Skip0
            
            Stack: nil
            
         meth initOutput(printInfo:  PrintInfo  <= @printInfo
                         printError: PrintError <= @printError
                         lockInfo:   LockInfo   <= @lockInfo
                         lockError:  LockError  <= @lockError)
            
            printInfo  <- PrintInfo
            printError <- PrintError
            
            lockInfo   <- LockInfo
            lockError  <- LockError
            
            self.printInfo = proc {$ VS}  Output, printInfo(VS) end
            self.showInfo  = proc {$ VS}  Output, showInfo(VS) end
            self.print     = proc {$ Val} Output, print(Val) end
            self.show      = proc {$ Val} Output, show(Val) end
            
            self.printError = proc {$ VS}  Output, printError(VS) end
            self.showError  = proc {$ VS}  Output, showError(VS) end
            self.printE     = proc {$ Val} Output, printE(Val) end
            self.showE      = proc {$ Val} Output, showE(Val) end
         end
         meth pushOutput(printInfo:  PrintInfo  <= @printInfo
                         printError: PrintError <= @printError
                         lockInfo:   LockInfo   <= @lockInfo
                         lockError:  LockError  <= @lockError
                         openInfo:   OpenInfo   <= Skip0
                         openError:  OpenError  <= Skip0
                         closeInfo:  CloseInfo  <= Skip0
                         closeError: CloseError <= Skip0)
            Top
         in
            {OpenInfo}
            {OpenError}
            
            Top = saved(printInfo:  @printInfo
                        printError: @printError
                        lockInfo:   @lockInfo
                        lockError:  @lockError
                        closeInfo:  @closeInfo
                        closeError: @closeError)

            Stack <- Top|@Stack
            
            printInfo  <- PrintInfo
            printError <- PrintError
            lockInfo   <- LockInfo
            lockError  <- LockError
            closeInfo  <- CloseInfo
            closeError <- CloseError
         end
         meth popOutput
            case @Stack of nil then
               raise error(output('stack::empty')) end
            [] Top|Rest then
               {@closeInfo}
               {@closeError}
               
               printInfo  <- Top.printInfo
               printError <- Top.printError
               lockInfo   <- Top.lockInfo
               lockError  <- Top.lockError
               closeInfo  <- Top.closeInfo
               closeError <- Top.closeError

               Stack <- Rest
            end
         end
         meth redirectOutput(Proc0
                             printInfo:  PrintInfo  <= @printInfo
                             printError: PrintError <= @printError
                             lockInfo:   LockInfo   <= @lockInfo
                             lockError:  LockError  <= @lockError
                             openInfo:   OpenInfo   <= Skip0
                             openError:  OpenError  <= Skip0
                             closeInfo:  CloseInfo  <= Skip0
                             closeError: CloseError <= Skip0)
            
            Output, pushOutput(printInfo:  PrintInfo 
                               printError: PrintError
                               lockInfo:   LockInfo
                               lockError:  LockError
                               openInfo:   OpenInfo   
                               openError:  OpenError 
                               closeInfo:  CloseInfo 
                               closeError: CloseError)
            try {Proc0}
            finally Output, popOutput end
         end
         meth redirectOutputToFile(Proc0
                                   stdout: FOut <= unit
                                   stderr: FErr <= unit)
            OutFile OutFlags StdOut
            ErrFile ErrFlags StdErr
            proc {OpenInfo}
               StdOut = {New Open.file init(name:OutFile flags:OutFlags)}
            end   
            proc {OpenError}
               StdErr = {New Open.file init(name:ErrFile flags:ErrFlags)}
            end
            proc {PrintInfo  VS} {StdOut write(vs:VS)} end
            proc {PrintError VS} {StdErr write(vs:VS)} end
            proc {CloseInfo}  {StdOut close} end
            proc {CloseError} {StdErr close} end
         in
            case FOut
            of append(F)   then OutFile=F     OutFlags=[write create append]
            [] truncate(F) then OutFile=F     OutFlags=[write create truncate]
            [] unit        then OutFile=false OutFlags=nil
            else                OutFile=FOut  OutFlags=[write create append]
            end
            case FErr
            of append(F)   then ErrFile=F     ErrFlags=[write create append]
            [] truncate(F) then ErrFile=F     ErrFlags=[write create truncate]
            [] unit        then ErrFile=false ErrFlags=nil
            else                ErrFile=FErr  ErrFlags=[write create append]
            end
            case (OutFile\=false)#(ErrFile\=false)
            of true#true then
               if {VS2S OutFile} \= {VS2S ErrFile} then
                  {self redirectOutput(Proc0
                                       printInfo:  PrintInfo
                                       printError: PrintError
                                       lockInfo:   {NewLock}
                                       lockError:  {NewLock}
                                       openInfo:   OpenInfo
                                       openError:  OpenError
                                       closeInfo:  CloseInfo
                                       closeError: CloseError)}
               else
                  SharedLock = {NewLock} in
                  {self redirectOutput(Proc0
                                       printInfo:  PrintInfo
                                       printError: PrintInfo
                                       lockInfo:   SharedLock
                                       lockError:  SharedLock
                                       openInfo:   OpenInfo
                                       closeInfo:  CloseInfo)}
               end
            [] true#false then
               {self redirectOutput(Proc0
                                    printInfo: PrintInfo
                                    lockInfo:  {NewLock}
                                    openInfo:  OpenInfo
                                    closeInfo: CloseInfo)}
            [] false#true then
               {self redirectOutput(Proc0
                                    printError: PrintError
                                    lockError:  {NewLock}
                                    openError:  OpenError
                                    closeError: CloseError)}
            [] false#false then
               {Proc0}
            end   
         end

         meth lockInfo($)
            @lockInfo
         end
         meth lockError($)
            @lockError
         end
         meth printInfo(VS)
            BS = {ByteString.make VS}
         in
            lock @lockInfo then
               {@printInfo BS}
            end
         end
         meth showInfo(VS)
            Output, printInfo(VS#"\n")
         end
         meth print(Val depth:Depth<=5 width:Width<=20)
            Output, printInfo({V2VS Val Depth Width})
         end
         meth show(Val depth:Depth<=5 width:Width<=20)
            Output, showInfo({V2VS Val Depth Width})
         end

         meth printError(VS)
            lock @lockError then
               {@printError VS}
            end
         end
         meth showError(VS)
            Output, printError(VS#"\n")
         end 
         meth printE(Val depth:Depth<=5 width:Width<=20)
            Output, printError({V2VS Val Depth Width})
         end
         meth showE(Val depth:Depth<=5 width:Width<=20)
            Output, showError({V2VS Val Depth Width})
         end
         meth printExc(Exc)
            {self printError({ExceptionToVirtualString Exc})}
         end
      end
      
      class Debug

         feat
            printDebug showDebug printD showD
            
         attr
            debug:      false
            printDebug: System.printInfo
            
         meth initDebug(debug:      D          <= @debug
                        printDebug: PrintDebug <= @printInfo)
            
            debug       <- D 
            printDebug  <- PrintDebug
            
            if {IsDet self.printDebug}==false then
               self.printDebug = proc {$ VS}  {self printDebug(VS)} end
               self.showDebug  = proc {$ VS}  {self showDebug(VS)} end
               self.printD     = proc {$ Val} {self printD(Val)} end
               self.showD      = proc {$ Val} {self showD(Val)} end
            end
         end

         meth debugStatus($)
            @debug
         end
         meth debugOn
            debug <- true
         end
         meth debugOff
            debug <- false
         end
         
         meth printDebug(VS)
            if @debug then
               {@printDebug VS}
            end
         end
         meth showDebug(VS)
            if @debug then
               Debug, printDebug(VS#"\n")
            end
         end 
         meth printD(Val depth:Depth<=5 width:Width<=20)
            if @debug then
               Debug, printDebug({V2VS Val Depth Width})
            end
         end
         meth showD(Val depth:Depth<=5 width:Width<=20)
            if @debug then
               Debug, showDebug({V2VS Val Depth Width})
            end
         end
      end

   end

   %%
   %% application args 
   %%
  
   local
      ArgsLock = {NewLock}
   in
      proc {ParseCmdArgs Spec Args ?ArgsRec}
         lock ArgsLock then 
            case {Property.condGet 'ozd.args' false} of false
            then
               AppArgs = {Property.get 'application.args'} in 
               {Property.put 'ozd.args' {Map Args VirtualString.toString}}
               {Application.getCmdArgs Spec ?ArgsRec}
               {Property.put 'ozd.args' {Map AppArgs VirtualString.toString}}
            elseof OzdArgs then
               {Property.put 'ozd.args' {Map Args VirtualString.toString}}
               {Application.getCmdArgs Spec ?ArgsRec}
               {Property.put 'ozd.args' {Map OzdArgs VirtualString.toString}}
            end
         end
      end
      proc {ParseCgiArgs Spec QueryString ?ArgsRec}
         lock ArgsLock then
            {OS.putEnv "REQUEST_METHOD" "GET"}
            {OS.putEnv "QUERY_STRING" QueryString}
            {Application.getCgiArgs Spec ?ArgsRec}
         end
      end
      
      class Args

         feat
            getCmdArgs
            
         attr
            args: nil
            
         meth initArgs(Args<=nil)
            args <- Args
            
            if {IsFree self.getCmdArgs} then
               self.getCmdArgs = fun {$ Spec} {self getCmdArgs(Spec $)} end
            end
         end
         
         meth getCmdArgs(Spec $)
            {ParseCmdArgs Spec @args}
         end
      end
   end
   
   %%
   %% application manager
   %%

   local 

      fun {FindThread L}
         case L of nil then raise xApplication(threadNotFound) end
         [] X|Rest then
            if X.'thread' == {Thread.this}
            then X
            else {FindThread Rest}
            end
         end
      end
         
      CondSelect = Value.condSelect

      fun {FindFeature X Feat}
         if {HasFeature X Feat}
         then X.Feat
         else {FindFeature X.parent Feat}
         end
      end
      
   in
      
      class ApplicationManager

         from
            Args Threads LocalEnv EnvAccess Output Debug
            
         feat
            args url info 'functor'
            
         meth init
            if {Property.condGet 'x-application.children' false} == false
            then
               {Property.put 'x-application.children' nil}
               
               self.url       = {Property.get 'application.url'}
               self.'functor' = false
               self.args      = {Property.get 'application.args'}
               self.info      = self.url#' (toplevel)'
               
               Args,      initArgs(self.args)
               Threads,   initThreads(exit:Application.exit)
               
               LocalEnv,  initLocalEnv
               EnvAccess, initEnvAccess
               Output,    initOutput
               Debug,     initDebug
            else
               X = {FindThread {Property.get 'x-application.children'}}
               {X.doFirst self}
               P     = X.parent
               ThrID = X.threadID
               proc {ExitProc ExitStatus}
                  thread {P tryKill(ThrID exit:ExitStatus)} end
                  {Thread.suspend {Thread.this}}
               end
            in
               self.url       = X.url
               self.'functor' = X.'functor'
               self.args      = X.args
               self.info      = X.info
             
               Args,    initArgs(self.args)
               Threads, initThreads(ThrID
                                    parent: P
                                    exit:   ExitProc
                                    modMan: {CondSelect X modMan P.modMan})
               if {HasFeature X env}
               then LocalEnv, initLocalEnv(env: X.env)
               else LocalEnv, initLocalEnv(inheritsFrom:{CondSelect X parentEnv
                                                         {FindFeature P env}})
               end
               EnvAccess, initEnvAccess(
                             getEnv:{CondSelect X getEnv
                                     {FindFeature P getLocalEnv}}
                             putEnv:{CondSelect X putEnv
                                     {FindFeature P putLocalEnv}})
               Output,    initOutput(
                             printInfo: {CondSelect X printInfo
                                         {FindFeature P printInfo}}
                             printError:{CondSelect X printError
                                         {FindFeature P printError}})
               Debug,     initDebug(
                             debug:     {CondSelect X debug
                                         {P debugStatus($)}}
                             printDebug:{CondSelect X printDebug
                                         {FindFeature P printDebug}})
            end
         end
         
         meth replicate
            {self.parent startApplication(self.url
                                          'functor': self.'functor'
                                          args:      self.args
                                          modMan:    self.modMan
                                          info:      self.info)}
         end
      end
   end

   %%
   %% wrapper class (for the XApplication.manager object)  
   %%

   %% The only reason for this construction is the following: When writing
   %% an application functor, you often want to print out error messages or
   %% something else at the very beginning of your code. By that time, the
   %% main object usually has not been initialized yet. Thus if the main object
   %% was just inheriting from the XApplication.manager, you would have no
   %% possibility to produce any output at all :-(
   
   class Wrapper
      
      feat
         appMan
         
      meth init(AppMan)
         self.appMan = AppMan
      end
      meth object($)
         self.appMan
      end
      
      meth printInfo(...)  = M  {self.appMan M}  end
      meth showInfo(...)   = M  {self.appMan M}  end
      meth print(...)      = M  {self.appMan M}  end
      meth show(...)       = M  {self.appMan M}  end
      
      meth printError(...) = M  {self.appMan M}  end
      meth showError(...)  = M  {self.appMan M}  end 
      meth printE(...)     = M  {self.appMan M}  end
      meth showE(...)      = M  {self.appMan M}  end

      meth printDebug(...) = M  {self.appMan M}  end
      meth showDebug(...)  = M  {self.appMan M}  end
      meth printD(...)     = M  {self.appMan M}  end
      meth showD(...)      = M  {self.appMan M}  end

      meth printExc(...)   = M  {self.appMan M}  end
      
      meth otherwise(M)         {self.appMan M}  end
   end

end


