%%%
%%% extensions to module Open.oz
%%%

functor
   
import              
   Open(socket:Socket pipe:Pipe text:Text) OS

   Process
   at 'x-ozlib://mathweb/share/Process.ozf'

   XCompiler
   at 'x-ozlib://mathweb/share/XCompiler.ozf'
   
export
   wrapper:    Wrapper
   socket:     XSocket
   server:     XServer
   client:     XClient
   msgChannel: MsgChannel
   pipe:       XPipe
   testPort:   TestPort

define

   class Wrapper
      attr RealObj
      meth init(Obj)
	 RealObj <- Obj
      end
      %% File
      meth read(...)     =M {@RealObj M} end
      meth write(vs:VS)  =M {@RealObj M} end
      meth seek(...)     =M {@RealObj M} end
      meth tell(...)     =M {@RealObj M} end
      meth close(...)    =M {@RealObj M} end
      %% additional methods for Socket & Pipe
      meth accept(...)   =M {@RealObj M} end
      meth receive(...)  =M {@RealObj M} end
      meth send(...)     =M {@RealObj M} end
      meth shutDown(...) =M {@RealObj M} end
      meth flush(...)    =M {@RealObj M} end
   end
   
   fun {TestPort P}
      S = {New Open.socket init} in
      try
         C in
	 {S bind(takePort:P)}
	 {S listen}
         C = {New Open.socket init}
	 {C connect(host:localhost port:P)}
	 {S accept}
	 {C shutDown} {C close}
	 {S shutDown} {S close}
	 true
      catch _ then false end
   end


   proc {Split S StopChar ?S1 ?S2}
      case S of C|Sr then
         if C==StopChar
         then S1=nil S2=Sr
         else S1=C|{Split Sr StopChar $ ?S2}
         end
      [] nil then
         S1 = S2 = S 
         raise connectionClosed end
      end
   end
   proc {Split0 S ?S1 ?S2}
      case S of C|Sr then
         if C==0
         then S1=nil S2=Sr
         else S1=C|{Split0 Sr $ ?S2}
         end
      [] nil then
         S1 = S2 = S
         raise connectionClosed end
      end
   end
   proc {Split128 S ?S1 ?S2}
      case S of C|Sr then
         if C==128
         then S1=nil S2=Sr
         else S1=C|{Split128 Sr $ ?S2}
         end
      [] nil then
         S1 = S2 = S
         raise connectionClosed end
      end
   end
   
   class XSocket
      from Socket
      prop locking

      feat
         ReadStopC
         WriteAppendVS
         SplitP
         ReaderThread
         Terminated
         
      attr
         list: unit
         tail: unit
         
      meth init(...) = M
         Socket, init
	 {self {Adjoin M doInit}}
      end

      meth doInit(...) = M
	 {self {Adjoin M set}}
	 {self startReader}
      end
      
      meth set(readStopC:     StopChar <= &\n
	       writeAppendVS: AppendVS <= [StopChar])
	 self.ReadStopC     = StopChar
	 self.WriteAppendVS = AppendVS
	 self.SplitP = case StopChar
		       of 0 then Split0
		       [] 128 then Split128
		       else proc {$ S ?S1 ?S2} {Split S StopChar ?S1 ?S2} end
                       end
	 local T in
	    list <- T
	    tail <- T
	 end
      end
      
      meth startReader
         thread
            self.ReaderThread={Thread.this}
            proc {ReadLoop}
               L T in
               Socket,read(list:@tail tail:T len:?L)
               if L==0
               then {Thread.terminate {Thread.this}}
               else tail<-T {ReadLoop}
               end
            end
         in
            try
               {Delay 100} % I don't know why, but this seems to be neccessary
	       {ReadLoop}
	    catch system(os(os "read" 131 ...)...) then 
	       skip % connection reset by peer
	    [] system(os(os "read" 104 ...)...) then %% added by jzimmer 17.9.2001
               skip                                  %% occured while leaving TPS service 
	                                             %% from within OMEGA
	    finally
               @tail=nil
               self.Terminated=unit
            end
         end
      end

      meth putS(S)
         Socket,write(vs:S#self.WriteAppendVS)
      end
      meth getS(?S)
         lock list<-{self.SplitP @list ?S $} end
      end
      meth getBS($)
         {ByteString.make XSocket,getS($)}
      end
      meth getInt($)
         {StringToInt XSocket,getS($)}
      end
      meth close
         try
            if {IsFree self.ReaderThread} then
               try self.ReaderThread=unit
               catch failure(...) then skip end
            end
            if self.ReaderThread\=unit then
               {Thread.terminate self.ReaderThread}
               {Wait self.Terminated}
            end
         catch error(kernel(deadThread...)...) then skip 
         finally
            try Socket, shutDown
            catch system(os(os "shutdown"...)...) then skip end
            Socket, close
         end
      end
   end

   class XServer
      from Socket XSocket

      meth init(port:Port ...) = M
         Socket, init
         if {IsDet Port}
         then Socket, bind(takePort:Port)
         else Socket, bind(port:?Port)
         end
	 Socket, listen
	 XSocket, {Adjoin {Record.subtract M port} set}
      end

      meth accept(...) = M
	 Socket, M
	 XSocket, startReader
      end
      
      meth close
         XSocket, close
      end
   end

   class XClient
      from Socket XSocket
                       
      meth init(host:H port:P...) = M
         Socket, init
         Socket, connect(host:H port:P)
         XSocket, {Adjoin {Record.subtractList M [host port]} set}
         XSocket, startReader
      end
      
      meth close
         XSocket, close
      end
   end
   
   V2VS = Value.toVirtualString
   
   class MsgChannel
      from XServer

      feat
         Read
         Write
         
      meth doInit(read:  DoRead  <= {XCompiler.new}
                  write: DoWrite <= fun {$ V} {V2VS V 100000 100000} end
                  ...) = M
         self.Read  = DoRead
         self.Write = DoWrite
         XServer, {Record.subtract {Record.subtract M read} write}
      end
      
      meth getVal($)
         {self.Read XServer,getS($)}
      end
      meth putVal(V)
         XServer, putS({self.Write V}) 
      end
   end

   /*
   local
      MakeBS  = ByteString.make
      SubStr  = ByteString.slice
      FindPos = ByteString.strchr
      EndPos  = ByteString.length
      BS2S    = ByteString.toString
   in
      class TestSocket from Socket
         attr
            buffer: {MakeBS ""}
            pos:    false
            
         meth readTill(StopChar<=0 byteString:?BS)
            fun {GetVS OldVS}
               case Socket, read(list:$) of nil then
                  raise connectionClosed end
               elseof S then
                  NewBS = {MakeBS S} in
                  case {FindPos NewBS 0 StopChar} of false then
                     {GetVS OldVS#NewBS}
                  elseof Pos then
                     buffer <- NewBS
                     pos    <- if Pos+1=={EndPos @buffer} then false
                               else Pos+1 end
                     {MakeBS OldVS#{SubStr NewBS 0 Pos}}
                  end
               end
            end
         in
            if @pos\=false then
               case {FindPos @buffer @pos StopChar} of false then
                  BS = {GetVS {SubStr @buffer @pos {EndPos @buffer}}}
               elseof Pos then
                  BS = {SubStr @buffer @pos Pos}
                  pos <- if Pos+1=={EndPos @buffer} then false
                         else Pos+1 end
               end
            else
               BS = {GetVS ""}
            end
         end

         meth readBS($)
            TestSocket,readTill(0 byteString:$)
         end
         meth readInt($)
            {StringToInt {BS2S TestSocket,readTill(0 byteString:$)}}
         end
      end
   end
   */

   local

      fun {`Record.subtractList` R L}
         {Record.filterInd R fun {$ F _} {Member F L}==false end}
      end
      
      CloseMsg = "Process terminated."

   in

      class XPipe
         from Text Pipe
            
         feat
            pid
            readDesc
            
	 meth init(cmd:Cmd args:Args...) = Init
	    self.pid      = Pipe, init(cmd:Cmd args:Args pid:$)
            self.readDesc = Pipe,getDesc($ _)
            thread
               {self {Adjoin {`Record.subtractList` Init [cmd args]} listen}}
            end
         end
         
         meth listen(showInfo: ShowInfo <= proc {$ _} skip end
                     onExit:   ExitProc <= proc {$} {ShowInfo CloseMsg} end)
            proc {Loop}
               case Text,getS($) of false then {ExitProc}
               [] Line then {ShowInfo Line} {Loop} end
            end
         in
            {OS.readSelect self.readDesc}
            {Loop}
         end
         
         meth close(kill:Kill<=false)
            {OS.deSelect self.readDesc}
            if Kill andthen {IsDet self.pid} then
               %%{OS.system 'kill -9 '#self.pid _}
               {Process.killTree self.pid}
            end
            thread Pipe,close end
            {Delay 600}
         end
      end

   end

end
