%%%
%%% Time
%%%

functor
   
import
   OS(localTime gmTime)
   Property(get)
   
export
   weekDay:        WeekDay         %% WeekDay.1 = 'Mon' 
   month:          Month           %% Month.10  = 'Nov'
   ord:            Ord             %% {Ord 1}='st', {Ord 2}='nd', {Ord 3}='rd' 
   
   mapToVS:        MapToVS         %% fun {MapToVS Rec} --> Rec1
   stamp:          Stamp           %% fun {Stamp Rec1}  --> VS
   
   localTime:      LocalTime       %% fun {LocalTime} 
   localTimeStamp: LocalTimeStamp  %% fun {LocalTimeStamp}
   localTimeHHMMSS:LocalTimeHHMMSS %% fun {LocalTimeHHMMSS}
   gmTime:         GMTime          %% fun {GMTime}
   gmTimeStamp:    GMTimeStamp     %% fun {GMTimeStamp}
   gmTimeHHMMSS:   GMTimeHHMMSS    %% fun {GMTimeHHMMSS}
   
   timer:    Timer
   %%           meth init   
   %%           meth start
   %%           meth getMSecs($)        
   %%           meth stop(MilliSeconds<=_)
   %%           meth get($)
   %%           meth reset

   watchDog: WatchDog
   %%           meth init(timeOut:  TimeOutMS
   %%                     onEvent:  EventProc <= proc {$} skip end
   %%                     onAlarm:  AlarmProc
   %%                     start:    Start     <= true)   
   %%           meth setTimeOut(TimeOut)     %
   %%           meth setEventProc(EventProc) %
   %%           meth setAlarmProc(AlarmProc) %
   %%           meth getTimeOut($)           %
   %%           meth getEventProc($)         %
   %%           meth getAlarmProc($)         %
   %%           meth start
   %%           meth event
   %%           meth stop

define
   WeekDay = wDay(0:'Sun' 1:'Mon' 2:'Tue' 3:'Wed' 4:'Thu' 5:'Fri' 6:'Sat')

   Month   = month(0:'Jan' 1:'Feb' 2:'Mar' 3:'Apr'  4:'May'  5:'Jun'
                   6:'Jul' 7:'Aug' 8:'Sep' 9:'Oct' 10:'Nov' 11:'Dec')

   fun {Ord N}
      if (N div 10) mod 10 == 1 then 'th'
      else
         case N mod 10
         of 1 then 'st'
         [] 2 then 'nd'
         [] 3 then 'rd'
         else 'th'
         end
      end
   end

   proc {Format2 N ?XY}
      _|XY = {IntToString 100+N}
   end

   fun {MapToVS T}
      WDay = WeekDay.(T.wDay)
      Mon  = Month.(T.mon)
      Day  = T.mDay
      XDay = {Ord Day}
      Year = 1900 + T.year
      Hour = {Format2 T.hour}
      Min  = {Format2 T.min}
      Sec  = {Format2 T.sec}
   in
      time(isDst:T.isDst
           wDay: WDay
           year: Year
           mon:  Mon
           day:  Day
           xDay: XDay
           hour: Hour
           min:  Min
           sec:  Sec)
   end

   fun {LocalTime} {MapToVS {OS.localTime}} end
   fun {GMTime}    {MapToVS {OS.gmTime}}    end
   
   fun {Stamp T}
      T.wDay#" "#T.mon#" "#T.day#T.xDay#" "#T.year#" "#
      T.hour#":"#T.min#":"#T.sec 
   end
   
   fun {LocalTimeStamp} {Stamp {LocalTime}} end
   fun {GMTimeStamp}    {Stamp {GMTime}}    end

   fun {HHMMSS T}
      T.hour#":"#T.min#":"#T.sec 
   end
   
   fun {LocalTimeHHMMSS} {HHMMSS {LocalTime}} end
   fun {GMTimeHHMMSS}    {HHMMSS {GMTime}}    end
   
   class Timer

      attr
         start: unit
         diff:  0

      meth init
         Timer, reset
      end

      meth start
         if @start == unit then
            start <- {Property.get 'time.total'}
         else
            raise timer(alreadyStarted) end
         end
      end         

      meth getMSecs($)
         if @start == unit
         then @diff
         else @diff + {Property.get 'time.total'} - @start
         end
      end
      
      meth stop(MilliSeconds<=_)
         if @start == unit then
            raise timer(notStarted) end
         else
            diff <- Timer, getMSecs($)
            start <- unit
         end
      end
      
      meth milliSec($)
         @diff
      end
      
      meth get($)
         MSecs = Timer, getMSecs($)
         Secs  = MSecs div 1000
         Mins  = Secs  div 60
         Hours = Mins  div 60
         Days  = Hours div 24
      in
         timeDiff(days:  Days
                  hours: Hours mod 24
                  mins:  Mins  mod 60
                  secs:  Secs  mod 60
                  msecs: MSecs mod 1000)
      end
      
      meth reset
         start <- unit
         diff <- 0
      end
      
   end

   class WatchDog
      prop
         locking
         
      attr
         timeOut
         eventProc
         alarmProc
         event
         'thread': unit
         
      meth init(timeOut:  TimeOutMS
                onEvent:  EventProc <= proc {$} skip end
                onAlarm:  AlarmProc
                start:    Start     <= true)
         WatchDog, setTimeOut(TimeOutMS)
         WatchDog, setEventProc(EventProc)
         WatchDog, setAlarmProc(AlarmProc)
         if Start then
            WatchDog, start
         end
      end
      meth setTimeOut(TimeOut)      timeOut   <- TimeOut    end
      meth setEventProc(EventProc)  eventProc <- EventProc  end
      meth setAlarmProc(AlarmProc)  alarmProc <- AlarmProc  end
      meth getTimeOut($)   @timeOut   end
      meth getEventProc($) @eventProc end
      meth getAlarmProc($) @alarmProc end
      
      meth start
         lock
            if @'thread'==unit then
               'thread' <- _
               thread
                  @'thread' = {Thread.this}
                  try WatchDog, Loop
                  finally 'thread'<-unit end
               end
               {Wait @'thread'}
            else
               raise watchDog(alreadyActive 'thread':@'thread') end
            end
         end
      end
      meth Loop
         E in
         event <- E
         {WaitOr E {Alarm @timeOut}}
         if {IsDet E} then
            {@eventProc}
            WatchDog, Loop
         else
            {@alarmProc}
         end
      end
      meth event
         @event=unit
      end   
      meth stop
         lock
            if @'thread'\=unit then
               try {Thread.terminate @'thread'}
               catch _ then skip end
            end
         end
      end
   end
end






