%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% %%
%%                                                                          %%
%%   Copyright (C) 1998 by AG Siekmann, Fachbereich Informatik,             %%
%%   Universitaet des Saarlandes, Saarbruecken, Germany.                    %%
%%   All rights reserved.                                                   %%
%%                                                                          %%
%%   For information about this program, write to:                          %%
%%     Stephan M. Hess                                                      %%
%%     AG Siekmann/FB Informatik                                            %%
%%     Universitaet des Saarlandes                                          %%
%%     Postfach 1150                                                        %%
%%     D-66041 Saarbruecken                                                 %%
%%     Germany                                                              %%
%%   electronic mail: hess@ags.uni-sb.de                                    %%
%%                                                                          %%
%%   The author makes no representations about the suitability of this      %%
%%   software for any purpose.  It is provided "AS IS" without express or   %%
%%   implied warranty.  In particular, it must be understood that this      %%
%%   software is an experimental version, and is not suitable for use in    %%
%%   any safety-critical application, and the author denies a license for   %%
%%   such use.                                                              %%
%%                                                                          %%
%%   You may use, copy, modify and distribute this software for any         %%
%%   noncommercial and non-safety-critical purpose.  Use of this software   %%
%%   in a commercial product is not included under this license.  You must  %%
%%   maintain this copyright statement in all copies of this software that  %%
%%   you modify or distribute.                                              %%
%%                                                                          %%
%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% %%

functor
   
import
   Tk
   TkTools
   Property
   
export
   textlistbox     : Textlistbox
   dropentry       : Dropentry
   flashlabel      : Flashlabel
   createImages    : CreateImages
   animlabel       : Animlabel
   help            : Help
   helpButton      : HelpButton
   helpButtonText  : HelpButtonText 
   textBrowser     : TextBrowser
   format_Frame    : Format_Frame
define

   %% from configure.oz:
   DefaultBorder     = 1
   DefaultPopupColor = 'lightyellow'
   DefaultStatusFont = case {Property.get 'platform.os'}
                       of win32 then default
                       else '-*-helvetica-medium-r-normal--*-100-*-*-*-*-*-*'
                       end
   DefaultHelpDelay  = 450
   EntryColor        = 'wheat'
   BrowserNormalFont = 'louistd'
   ScrollerWidth     = 12 
   local
      fun {StringLess Is Js}
         case Is of nil then true
         [] I|Ir then
            case Js of nil then false
            [] J|Jr then I<J orelse I==J andthen {StringLess Ir Jr}
            end
         end
      end
   in
      class Textlistbox
         from TkTools.textframe
               
         feat
            listbox
               
         meth tkInit(title        : Title         <= 'Listbox'
                     parent       : Parent
                     height       : Height        <= 10
                     width        : Width         <= 20
                     scrollerwidth: Scrollerwidth <= 15
                     buttons      : Buttons       <= nil
                     bd           : Border        <= DefaultBorder
                     bg           : Color         <= white
                     font         : Font          <= default)
            
            TkTools.textframe,tkInit(parent: Parent
                                     text  : Title)

            ListboxFrame = {New Tk.frame tkInit(parent: self.inner
                                                relief: flat)}
	 
            Listbox   = case Font of default then
                           {New Tk.listbox tkInit(parent: ListboxFrame
                                                  height: Height
                                                  width : Width
                                                  bg    : Color)}
                        else
                           {New Tk.listbox tkInit(parent: ListboxFrame
                                                  height: Height
                                                  width : Width
                                                  bg    : Color
                                                  font  : Font)}
                        end
	 
            Scrollbar = {New Tk.scrollbar tkInit(parent: ListboxFrame
                                                 bd    : Border
                                                 width : Scrollerwidth)}

            ButtonFrame = {New Tk.frame tkInit(parent: self.inner
                                               relief: flat)}

            ButtonList = {Map Buttons
                          fun {$ B}
                             {New Tk.{Label B} tkInit(parent: ButtonFrame
                                                      text  : B.text
                                                      action: B.action)}
                          end}
         in
            self.listbox = Listbox
            {Tk.addYScrollbar Listbox Scrollbar}
            {Tk.batch [pack(Listbox Scrollbar side:left fill:y)
                       case ButtonList of nil then
                          unit
                       else
                          pack(b(ButtonList) side:top fill:x)
                       end
                       pack(ListboxFrame ButtonFrame side:left fill:y)]}
         end

         meth tkBind(event : Event
                     action: Action)
            {self.listbox tkBind(event : Event
                                 action: Action)}
         end
	 
         meth getSelectedPosition(?Position)
            Index = {self.listbox tkReturn(curselection $)}
         in
            if Index\=nil then
               Position = Index
            else
               Position = false
            end
         end

         meth getSelection(?Element)
            Index = {self.listbox tkReturn(curselection $)}
         in
            if Index\=nil then
               Element = {self.listbox tkReturn(get(Index) $)}
            else
               Element = false
            end
         end

         meth clearSelection
            Index = {self.listbox tkReturn(curselection $)}
         in
            if Index\=nil then
               {self.listbox tk(selection clear Index)}
            else
               skip
            end
         end

         meth removeSelection(?Element)
            Index = {self.listbox tkReturn(curselection $)}
         in
            if Index\=nil then
               Element = {self.listbox tkReturn(get(Index) $)}
               Textlistbox,deleteSelection
            else
               Element = false
            end
         end
	 
         meth deleteSelection
            Index = {self.listbox tkReturn(curselection $)}
         in
            if Index\=nil then
               {self.listbox tk(delete Index)}
            else
               skip
            end
         end

         meth seeSelection
            Index = {self.listbox tkReturn(curselection $)}
         in
            if Index\=nil then
               {self.listbox tk(see Index)}
            else
               skip
            end
         end
	 
         meth insertElement(String Index <= 'lexicographical')
	    if Index=='lexicographical' then
	       Elements = {self.listbox tkReturnList(get('0' 'end') $)}
	       proc {InsertLexi Es S I}
		  case Es of nil then
		     {self.listbox tk(insert I S)}
		  [] E|Er then
		     if {StringLess S E} then
			{self.listbox tk(insert I S)}
		     else
			{InsertLexi Er S I+1}
                     end
		  end
	       end
	    in
               {InsertLexi Elements String 0}
            else
               {self.listbox tk(insert Index String)}
            end
         end

	 meth insertElements(Es)
            {ForAll Es proc{$ E}
                          Textlistbox,insertElement(E)
                       end}
         end

         meth insertElementsInOrder(Es)
	    {List.forAllInd Es proc {$ I E}
				  Textlistbox,insertElement(E I)
			       end}
         end
      
         meth getElements(?Elements)
            Elements = {self.listbox tkReturnList(get('0' 'end') $)}
         end
	 
         meth removeElements(?Elements)
            Textlistbox,getElements(Elements)
            {self.listbox tk(delete '0' 'end')}
         end

         meth clearElements
            {self.listbox tk(delete '0' 'end')}
         end
      
         meth emptyP($)
            {self.listbox tkReturnList(get('0' 'end') $)}==nil
         end

      end % of class
   
   local
      
      MaxEntries = 40
         
         class PopupListbox
            from Tk.toplevel
               
            feat
               listbox
	    
            meth tkInit(width    : Width
                        height   : Height
                        xposition: X
                        yposition: Y
                        bg       : Background)

               Tk.toplevel,tkInit(withdraw: true)
	 
               Listbox  = {New Tk.listbox tkInit(parent: self
                                                 width : Width
                                                 height: Height
                                                 bg    : Background)}
            in
               self.listbox = Listbox

               {Listbox tk(conf cursor: top_left_arrow)}

               {Tk.batch [wm(overrideredirect self true)
                          wm(geometry self '+'#X#'+'#Y)
                          pack(Listbox)
                          wm(deiconify self)]}
            end
      
            meth tkBind(event : Event
                        action: Action)
               {self.listbox tkBind(event : Event
                                    action: Action)}
            end
 
            meth setSelection(Index)
               {self.listbox tk(selection set Index)}
            end
      
            meth getSelection(?Element)
               Index = {self.listbox tkReturn(curselection $)}
            in
               if Index\=nil then
                  Element = {self.listbox tkReturn(get(Index) $)}
               else
                  Element = false
               end
            end
      
            meth seeSelection
               Index = {self.listbox tkReturn(curselection $)}
            in
               if Index\=nil then
                  {self.listbox tk(see Index)}
               else
                  skip
               end
            end
	 
            meth insertElementAndSet(String Set Index <= 'lexicographical')
               if Index=='lexicographical' then
                  Elements = {self.listbox tkReturnList(get('0' 'end') $)}
                  proc {InsertLexi Es S C I}
                     case Es of nil then
                        {self.listbox tk(insert I S)}
                        if S==C then
                           PopupListbox,setSelection(I)
                        else skip end
                     [] E|Er then
                        if {StringLess S E} then
                           {self.listbox tk(insert I S)}
                           if S==C then
                              PopupListbox,setSelection(I)
                           else skip end
                        else
                           {InsertLexi Er S C I+1}
                        end
                     end
                  end
               in
                  {InsertLexi Elements String Set 0}
               else
                  {self.listbox tk(insert Index String)}
                  if String==Set then
                     PopupListbox,setSelection(Index)
                  else skip end
               end
            end
      
            meth insertElementsAndSet(Es S)
               {ForAll Es proc{$ E}
                             PopupListbox,insertElementAndSet(E S)
                          end}
               PopupListbox,seeSelection
            end
      
         end %of class

      in
   
         class Dropentry
            from Tk.frame

            attr
               background
               elements

            feat
               'do'
               entry
               width
               dropButton
	    
            meth tkInit(parent    : Parent
                        'do'      : Procedure  <= proc {$ _} skip end
                        elements  : Elements   <= nil
                        selection : Selection  <= nil
                        width     : Width      <= 10
                        border    : Border     <= 2
                        bg        : Background  <= EntryColor
                        dropGif   : DropGif)
	 
               Tk.frame,tkInit(parent: Parent)
	 
               Entry = {New Tk.entry tkInit(parent: self
                                            width : Width
                                            border: Border
                                            bg    : Background)}

               DropImage = {New Tk.image tkInit(type  : photo
                                                format: gif
                                                url   : DropGif)}
	 
               DropButton = {New Tk.button tkInit(parent: self
                                                  relief: flat
                                                  bd    : 1
                                                  image : DropImage
                                                  action: self # DropListbox)}
            in
               self.'do'       = Procedure
               self.entry      = Entry
               self.width      = Width
               self.dropButton = DropButton
               background     <- Background
               elements       <- Elements

               {Entry tk(insert 0 Selection)}
               {Entry tk(conf state:disabled)}

               {Tk.send pack(Entry DropButton side:left)}
            end

            meth DropListbox
               [X Y H] = {Map [rootx rooty height]
                          fun {$ WI}
                             {Tk.returnInt winfo(WI self.entry)}
                          end}

               Listbox = {New PopupListbox tkInit(width    : self.width
                                                  height   : {Min {Length @elements} MaxEntries}
                                                  xposition: X
                                                  yposition: Y+H-2
                                                  bg       : @background)}
            in
               Dropentry,DisableDrop
               {Listbox tkBind(event : '<ButtonRelease-1>'
                               action: proc {$}
                                          S = {Listbox getSelection($)}
                                       in
                                          {self.'do' S}
                                          {self insert(S)}
                                          {Listbox tkClose}
                                          {self EnableDrop}
                                       end)} 
               {Listbox insertElementsAndSet(@elements {self get($)})}
            end

            meth addElement(Element)
               elements <- Element|@elements
            end
	 
            meth setElements(Elements)
               elements <- Elements
            end
	 
            meth insert(String)
               {self.entry tk(conf state:normal)}
               {self.entry tk(delete 0 'end')}
               {self.entry tk(insert 0 String)}
               {self.entry tk(conf state:disabled)}
            end
      
            meth get(?String)
               {self.entry tkReturn(get ?String)}
            end

            meth EnableDrop
               {self.dropButton tk(configure state: normal)}
            end

            meth DisableDrop
               {self.dropButton tk(configure state: disabled)}
            end
	 
         end %of class

      end %of local
      class Flashlabel
         from
            Tk.label
            Time.repeat
         
         feat
            text
            color

         attr
            visible: true
	 
         meth init(parent: Parent
                   delay : Delay <= 1000)

            Tk.label,tkInit(parent: Parent)

            Time.repeat,setRepAll(action: proc {$}
                                             {self FlashText}
                                          end
                                  delay : Delay)
         end

         meth start(text : Text
                    color: Color)
            Tk.label,tk(conf text: Text)
            self.text  = Text
            self.color = Color
            Flashlabel,go
         end
      
         meth FlashText
            if @visible then
               Tk.label,tk(conf fg: c(216 216 216))
               visible <- false
            else
               Tk.label,tk(conf fg: self.color)
               visible <- true
            end
         end

         meth go
            thread Time.repeat,go end
         end
      
         meth stop
            Time.repeat,stop
         end
      
      end %of class
   
      local
         proc {CreateIs BaseName Digits Suffix Current ?Images}
            local
               local
                  fun {MkZeroes N}
                     {ForThread 1 N 1 fun {$ X _} X#'0' end nil}    
                  end
               in
                  fun {ExtendNumber N Digits}
                     S = {Int.toString N} in
                     {MkZeroes Digits-{Length S}}#S
                  end
               end
            in
               URL = BaseName#{ExtendNumber Current Digits}#Suffix
            end
         in
            try
               Image = {New Tk.image tkInit(type:photo format:gif url:URL)}
            in
               Images = Image|{CreateIs BaseName Digits Suffix Current+1}
            catch _ then
               Images = nil
            end
         end
      in
         proc {CreateImages BaseName Digits Suffix ?Images}
            {CreateIs BaseName Digits Suffix 0 ?Images}
         end
      end
      
      local
         skip
      in   
         class Animlabel
            from Tk.label Time.repeat
	 
            prop
               final

            attr
               Current
               Maximum
	 
            meth init(parent   : Parent
                      images   : Is
                      framerate: FrameRate <= 50)
               
               Images = if {Label Is}=='|' then {List.toTuple '#' Is}
                        else Is end
            in
               Current <- 1
               Maximum <- {Width Images}
               Tk.label, tkInit(parent:Parent image:Images.1)
               local
                  proc {Action}
                     N = {self NextPos($)} in
                     Tk.label, tk(conf image:Images.N)
                  end
               in
                  Time.repeat, setRepAll(action: Action
                                         delay:  1000 div FrameRate)
               end
            end
            %%
            meth NextPos(?N)
               Pos = @Current + 1
            in
               N = if Pos > @Maximum
                   then 1
                   else Pos
                   end
               Current <- N
            end
            %%
            meth setFrameRate(Rate)
               Time.repeat, setRepDelay(1000 div Rate)
            end
            %%
            meth go
               thread Time.repeat, go end
            end
            %%
            meth stop
               Time.repeat, stop
            end
         end         
      end

      %%
      %% help classes
      %%   

      class Help
         attr
            Stamp :{NewName}
            Window:true
         
         feat
            Popup
            Message
            Status
            StartDelay
         
         meth help(help:Text message:M status:S
                   popupColor: PopupColor <= DefaultPopupColor
                   statusFont: StatusFont <= DefaultStatusFont
		   delay:      HelpDelay  <= DefaultHelpDelay)
	    
            {self tkBind(event:'<Enter>'  action:self#Start append:true)}
            {self tkBind(event:'<Leave>'  action:self#Stop  append:true)}
            {self tkBind(event:'<Button>' action:self#Stop  append:true)}
         
            self.Popup =
            fun {$ X Y}
               T = {New Tk.toplevel tkInit(withdraw:true bg:black)}
               M = {New Tk.message tkInit(parent:T font:StatusFont text:Text
                                          bg:PopupColor aspect:800)}
            in
               {Tk.batch [wm(overrideredirect T true)
                          wm(geometry T '+'#X#'+'#Y)
                          pack(M padx:2 pady:2)
                          wm(deiconify T)]}
               T
            end
            self.Message    = M
	    self.Status     = S
            self.StartDelay = HelpDelay
         end
      
         meth Start
            NewStamp={NewName}
         in
            Stamp<-NewStamp
            {Delay self.StartDelay} 
            {self Display(NewStamp)}
         end
      
         meth Stop
            Stamp<-{NewName} 
            if {IsObject @Window} then
               {@Window tkClose}
               {self.Status clrMessage}
            else
               skip
            end
         end
      
         meth Display(NewStamp)
	    if @Stamp==NewStamp then
	       [X Y W H] = {Map [rootx rooty width height]
			    fun {$ WI} {Tk.returnInt winfo(WI self)} end}
            in
	       Window <- {self.Popup X+W div 2 Y+H}
	       {self.Status setMessage(self.Message)}
            else skip
            end
         end
      
      end
   
      %%
      class HelpButton
         from Tk.button Help
         
         meth init(parent:Parent status:Status action:Action
                   file:File help:Help message:Message url:Url)
         
            Image = {New Tk.image tkInit(type  : photo
                                         format: gif
					 url   : Url # File)}
	 in
            HelpButton,tkInit(parent:Parent
			      borderwidth:1
			      state:normal
                              action:Action
			      image :Image)
	    HelpButton,help(help:Help message:Message status:Status)
         end    
      end
   
      %%
      class HelpButtonText
         from Tk.button Help
         
         meth init(parent:Parent status:Status action:Action text:Text
                   help:Help message:Message)
            HelpButtonText,tkInit(parent:Parent action:Action text:Text)
            HelpButtonText,help(help:Help message:Message status:Status)
         end
      end
   end
    class TextBrowser
      from Tk.frame         
      feat text
      meth init(p:Parent h:H w:W bg:BG<='white' tf:TF<=false title:Title<='')
         Tk.frame,tkInit(parent: Parent)
         TextFrame = if TF then {New TkTools.textframe tkInit(parent:self text:Title)}
                     else TextFrame = unit  end
         LocParent = if TextFrame \= unit then TextFrame.inner else self end 
         Text = {New Tk.text tkInit(parent: LocParent height: H width: W border: 0
                                    bg: BG font: BrowserNormalFont state: disabled wrap: none)}
         
         ScrX = {New Tk.scrollbar tkInit(parent:LocParent relief: sunken bd:DefaultBorder
                                         width: ScrollerWidth orient: horizontal)}
      
         ScrY = {New Tk.scrollbar tkInit(parent:LocParent relief: sunken bd: DefaultBorder
                                         width: ScrollerWidth)}
      in
         self.text = Text
         {Tk.addXScrollbar Text ScrX}
         {Tk.addYScrollbar Text ScrY}
         if TF then 
            {Tk.batch [grid(Text row:0 column:0 sticky:nswe)
                       grid(ScrY row:0 column:1 sticky:ns)
                       grid(ScrX row:1 column:0 sticky:we)
                       grid(columnconfigure TextFrame.inner 0 weight:1.0)
                       grid(rowconfigure    TextFrame.inner 0 weight:1.0)
                       pack(TextFrame)]}
         else
            {Tk.batch [grid(Text row:0 column:0 sticky:nswe)
                       grid(ScrY row:0 column:1 sticky:ns)
                       grid(ScrX row:1 column:0 sticky:we)]}
         end
      end
      
      meth clear
         {self.text tk(conf state:normal)}
         {self.text tk(delete '1.0' 'end')}
         {self.text tk(conf state:disabled)}
      end
      
      meth insert(Text)
         {self.text tk(conf state:normal)}
         {self.text tk(insert 'end' Text)}
         {self.text tk(yview 'moveto' 1.0)}
         {self.text tk(xview 'moveto' 0)}
         {self.text tk(conf state:disabled)}
      end      
   end %of clas
   local
      class Label_Column  
         from Tk.frame
            
         prop final
         feat label1 label2
            
         meth init(Parent L1 L2)
            Tk.frame,tkInit(parent: Parent border: 1 relief: flat)
            
            Label1 = {New Tk.label tkInit(parent: self text:L1)}
            Label2 = {New Tk.label tkInit(parent: self text:L2)}
         in
            self.label1 = Label1
            self.label2 = Label2
            {Tk.send pack(Label1 Label2 anchor:w side:top)}             
         end         
      end %of class
      class Entry_Column  
         from Tk.frame
            
         prop final
            
         feat entry1 entry2      
         meth init(Parent)
            Tk.frame,tkInit(parent: Parent border: 1 relief: flat)
            
            Entry1={New Tk.entry tkInit(parent: self width: 30 state: disabled border: DefaultBorder bg: white)}
            
            Entry2={New Tk.entry tkInit(parent: self width: 30 state: disabled border: DefaultBorder bg: white)}            
         in            
            {Entry1 tkBind(event : '<Return>'
                           action: proc {$} {Entry1 tk(conf state:disabled)} end 
                           append: true)}
            
            {Entry1 tkBind(event : '<Button-1>'
                           action: proc {$} {Entry1 tk(conf state:normal)} end
                           append: true)}
            
            self.entry1  = Entry1
            self.entry2  = Entry2
            {Tk.send pack(Entry1 Entry2 anchor:w side:top)} 
         end         
      end
   in
      class Format_Frame  
         from Tk.frame
            
         prop
            final
            
         feat  entries1 entries2 labels1 labels2
            
         meth init(Parent GenreName Title1 Title2 Title3)
            Tk.frame,tkInit(parent: Parent border: 1 relief: flat)
            
            Entries1 = {New Entry_Column init(self)}
            Labels1  = {New Label_Column init(self GenreName Title1)}	 
            Entries2 = {New Entry_Column init(self)}
            Labels2  = {New Label_Column init(self Title2 Title3)}
            
         in
            self.entries1 = Entries1
            self.labels1  = Labels1
            self.entries2 = Entries2
            self.labels2  = Labels2
            {Tk.send pack(Labels1 Entries1 Labels2 Entries2 side:left)} 
         end         
      end % of class      
   end

end
%%% Changes:

%%% PopupListbox, tkInit(...):
%%%   29.8.99 new optional parameter bg: EntryColor <= 'wheat'
%%% DropEntry,tkInit(...):
%%%  29.8.99 new parameter dropGif: DropGif  
%%% FlashLabel,go & AnimLabel,go:
%%%  29.8.99 {App newThread(...)} --> thread ... end
%%% Help, HelpButton, HelpButtonText:
%%%  14.9.99 added here, were in help.oz and in many other files














