%%% -*- indent-tabs-mode: nil -*-
%%%
%%% Author:
%%%   Juergen Zimmer <jzimmer@mathweb.org>
%%%
%%% Copyright:
%%%   Juergen Zimmer, 2001
%%%
%%% Last change:
%%%   $Date: 2001/09/13 15:37:46 $ by $Author: jzimmer $
%%%   $Revision: 1.2 $
%%%
%%% This file is part of MathWeb
%%%    http://www.mathweb.org/mathweb
%%%
%%% See the file "LICENSE" or
%%%    http://www.mathweb.org/mathweb/LICENSE.html
%%% for information on usage and redistribution
%%% of this file, and for a DISCLAIMER OF ALL
%%% WARRANTIES.
%%%

functor
   
import
   
   Type(ask)
%   Error(exceptionToMessage messageToVirtualString)
%   Property(put)
%   URL(make toVirtualString)
%   Browser % this is used only if Debug == true
%   System
   
%   HttpHandler('class':BaseHTTPHandler)
%   at 'x-ozlib://mathweb/httpd/HttpHandler.ozf'
   
%   XmlParser
%   at 'x-ozlib://mathweb/xmlparser/XmlParser.ozf'
%   PModule
%   at 'x-ozlib://mathweb/xmlrpc/xmlrpc.parser.ozf'
%   SModule
%   at 'x-ozlib://mathweb/xmlrpc/xmlrpc.sem.ozf'
%   XmlrpcWriter('class':Xmlrpc)
%   at 'x-ozlib://mathweb/xmlrpc/xmlrpc.printer.ozf'
   
export
   'class': XmlPrinter
   

define

   %% GetOptions is unchanged
   fun {GetOptions As N ?M}
      case As of nil then M=N nil
      [] A|Ar then
         if {IsInt A} then {GetOptions Ar N+1 M}
         else M=N As
         end
      end
   end      
   %% BuildOptions is unchanged
   fun {BuildOptions As Tag}
      case As of nil then ''
      [] A|Ar then ' '#A#'="'#Tag.A#'"'#{BuildOptions Ar Tag}
      end
   end

   local
      fun {NoInt X}
         {IsInt X}==false
      end
   in
      fun {AllAttribs Tag}
         {Filter {Arity Tag} NoInt}
      end
   end
   fun {Verb L}
      case L of nil then "nil" else L end
   end

   %% This class requires the existence of a write(vs:...) method.
   %% It must be mixed with one of Open.file/socket/pipe etc.
   %%
   %% the following is modified code from the Open module
   %%
   class XmlPrinter

      meth startTag(T attribs:Attribs<={AllAttribs T})
         L = {Label T} in
         {self write(vs:'<'#{Verb L}#{BuildOptions Attribs T}#'>')}
      end
      meth endTag(T)
         L = {Label T} in
         {self write(vs:'</'#{Verb L}#'>')}
      end
      meth emptyTag(T attribs:Attribs<={AllAttribs T})
         L = {Label T} in
         {self write(vs:'<'#{Verb L}#{BuildOptions Attribs T}#' />')}
      end
      %%
      meth tagBody(N Tag)
         {For 1 N 1
          proc {$ I} {self tag(Tag.I)} end}
      end
      meth tagTuple(Tag)
         case {Label Tag}
         of '#' then
            {Record.forAll Tag
             proc {$ T} {self tag(T)} end}
         [] '|' then
            {self write(vs: Tag)}
         [] L then
            case Tag %% new
            of 'self'(M) then {self M}
            [] ','(C M)  then C, M
            [] nil then skip
            [] '' then skip
            else
               W = {Width Tag} in
               if W>0 then
                  {self startTag(L)}
                  {self tagBody(W Tag)}
                  {self endTag(L)}
               else
                  {self emptyTag(L)}
               end
            end
         end
      end
      meth tagRecord(Tag)
         N As={GetOptions {Arity Tag} 0 ?N}
      in
         if N>0 then
            {self startTag(Tag attribs:As)}
            {self tagBody(N Tag)}
            {self endTag(Tag)}
         else
            {self emptyTag(Tag attribs:As)}
         end
      end
      meth tagProcedure(P)
         case {Procedure.arity P}
         of 0 then {P} %% new
         [] 1 then {self tag({P})}
         else % raise type error
            {Type.ask.'procedure/1' P}
         end
      end
      meth tag(Tag)
         if {IsTuple Tag} then
            XmlPrinter, tagTuple(Tag) 
         elseif {IsRecord Tag} then
            XmlPrinter, tagRecord(Tag)   
         elseif {IsProcedure Tag} then
            XmlPrinter, tagProcedure(Tag)
         else
            {self write(vs:Tag)}
         end
      end
   end

end % define
