%%% Original file:
%%%   Gif2Functor.oz
%%%
%%% Author:
%%%   Nils Franzn (nilsf@sics.se)
%%%
%%% Copyright:
%%%   Nils Franzn, 1999
%%%
%%% Last change:
%%%   $Date: 2001/07/30 21:35:20 $ by $Author: afranke $
%%%   $Revision: 1.1 $
%%%
%%% This file is part of Mozart, an implementation
%%% of Oz 3
%%%    http://www.mozart-oz.org
%%%
%%% See the file "LICENSE" or
%%%    http://www.mozart-oz.org/LICENSE.html
%%% for information on usage and redistribution
%%% of this file, and for a DISCLAIMER OF ALL
%%% WARRANTIES.
%%%

%%% this file (gif2data.oz) derived by
%%%   Andreas Franke (afranke@ags.uni-sb.de)

functor 
import
   Application(getCmdArgs exit) 
   Open(file) 
   Property(get) 
   System(printError showError)

   GifToBase64(convert) 

define
   Spec=record('output'(single char:&o type:string default:nil))

   local
      proc {SaveVS VS File}
         F = {New Open.file init(name:File flags:[write create truncate])}
      in
         {F write(vs:VS)}
         {F close}
      end
   in    
      proc {ConvertToFile InFile OutFile}
         {SaveVS {GifToBase64.convert InFile} OutFile}
      end
   end

   fun {DropGifExt FileNameVS}
      FileNameS = {VirtualString.toString FileNameVS}
   in
      if {Map {Reverse {List.take {Reverse FileNameS} 4}} Char.toLower}==".gif"
      then {List.take FileNameS {Length FileNameS}-4}
      else FileNameS
      end
   end

   proc {ProcessFile InFile OutFile}
      Out = if OutFile==nil then {DropGifExt InFile}#".data"
            else OutFile end
   in
      try
         {ConvertToFile InFile Out}
      catch system(os(...) ...) then
         {System.showError "Can't find input file: "#InFile}
         {Application.exit 2}
      end
   end

   Args 
   try
      Args = {Application.getCmdArgs Spec}
   catch error(ap(usage S)...) then
      AppURL={Property.get 'application.url'} in
      {System.printError AppURL#": "#S#"\n\n"#
       'Usage: '#AppURL#' <GIF-File> [-o <DATA-File>]\n'#
       '     | '#AppURL#' <GIF-Files> ... \n'}
      {Application.exit 2}
   end

   case Args.1 of nil then % No input
      {System.printError "No GIF file given\n"}
      {Application.exit 2}
   [] [F] then 
      {ProcessFile F Args.output}
   [] Files then
      if Args.output\=nil then
         AppURL={Property.get 'application.url'} in
         {System.printError AppURL#": "#
          "`-o' option not allowed with multiple input files!\n"}
         {Application.exit 2}
      else
         {ForAll Files proc {$ F} {ProcessFile F nil} end}
      end
   end
   {Application.exit 0}
end

/*

declare
proc {LoadS File ?S}
   F = {New Open.file init(name:File)}
in
   {F read(list:?S size:all)}
   {F close}
end
Data = {LoadS "/home/afranke/ozgifs/symbols/update.data"}
T={New Tk.toplevel tkInit()}
I={New Tk.image tkInit(type:photo format:gif data:Data)}
L={New Tk.label tkInit(parent:T image:I)}
{Tk.send pack(L)}

*/
