%%% Original file:
%%%   Gif2Functor.oz
%%%
%%% Author:
%%%   Nils Franzn (nilsf@sics.se)
%%%
%%% Copyright:
%%%   Nils Franzn, 1999
%%%
%%% Last change:
%%%   $Date: 2001/07/30 21:35:21 $ by $Author: afranke $
%%%   $Revision: 1.1 $
%%%
%%% This file is part of Mozart, an implementation
%%% of Oz 3
%%%    http://www.mozart-oz.org
%%%
%%% See the file "LICENSE" or
%%%    http://www.mozart-oz.org/LICENSE.html
%%% for information on usage and redistribution
%%% of this file, and for a DISCLAIMER OF ALL
%%% WARRANTIES.
%%%

%%% this file (gifs2functor.oz) derived by
%%%   Andreas Franke (afranke@ags.uni-sb.de)

functor 
import
   Browser(browse)
   Application(getCmdArgs exit)
   GifToBase64(convert) 
   System(show printError)
   Pickle
   Property(get)

   XCompiler(newWithEnv)
   at 'x-ozlib://mathweb/share/XCompiler.ozf'
   %% 'http://www.ags.uni-sb.de/~omegadev/mathweb/functors/XCompiler.ozf'
   %%  http://www.ags.uni-sb.de/~omegadev/mathweb/prog/share/XCompiler.oz
   
define
   Spec = record('output'(single char:&o type:string default:nil)
                 'compress'(single char:&z type:int(min:0 max:9) default:0))
   
   fun {GetGifData FileName}
      DataS = try
                 {GifToBase64.convert FileName}
              catch system(os(os "open" 2...)...) then
                 {GifToBase64.convert FileName#".gif"}
              end
   in
      {ByteString.make DataS}
   end
   
   fun {DropGifExt FileNameVS}
      FileNameS = {VirtualString.toString FileNameVS}
   in
      if {Map {Reverse {List.take {Reverse FileNameS} 4}} Char.toLower}==".gif"
      then {List.take FileNameS {Length FileNameS}-4}
      else FileNameS
      end
   end
   
   fun {Convert Fs}
      GifsAList = {Map Fs fun {$ F}
                             {StringToAtom {DropGifExt F}}#{GetGifData F}
                          end}
      GifsRec = {Record.adjoinList gifsData GifsAList}
      VS2V    = {XCompiler.newWithEnv env('Data':GifsRec)}
      
      fun {WriteExports L}
         case L of nil then ""
         [] X|R then
            "\n   '"#X#"': `"#X#"`"#{WriteExports R}
         end
      end
      fun {WriteDefines L}
         case L of nil then ""
         [] X|R then
            "\n   `"#X#"` = {ImgByNeed Data.'"#X#"'}"#{WriteDefines R}
         end
      end

      FunctorVS =
      ""   #"functor"
      #"\n"#"import"
      #"\n"#"   Tk(image)"
      #"\n"#"export "#{WriteExports {Arity GifsRec}}
      #"\n"#"define"
      #"\n"#"   fun {ImgByNeed BS}"
      #"\n"#"      {Value.byNeed"
      #"\n"#"       fun {$}"
      #"\n"#"          Data = {ByteString.toString BS} in"
      #"\n"#"          {New Tk.image tkInit(data:Data type:photo format:gif)}"
      #"\n"#"       end}"
      #"\n"#"   end"
      #"\n"#"   "#{WriteDefines {Arity GifsRec}}    
      #"\n"#"end "
      #"\n"
   in
      /*
      local
         proc {SaveVS VS File}
            F = {New Open.file init(name:File flags:[write create truncate])}
         in
            {F write(vs:VS)}
            {F close}
         end
      in
         {SaveVS FunctorVS "/tmp/functor.oz"}
      end
      */
      {VS2V FunctorVS}
   end
   
   proc{ConvertToFile Fs Out Compress}
      OutFile = if Out==nil then "Gifs.ozf"
                else Out end
   in
      {Pickle.saveCompressed {Convert Fs} OutFile Compress}
   end

   try
      Args={Application.getCmdArgs Spec}
   in
      case Args.1 of nil then % No input
         {System.printError "No GIF file given\n"}
         {Application.exit 2}
      [] GifFiles then % single file
         try
            {ConvertToFile GifFiles Args.output Args.compress}
            {Application.exit 0}
         catch Exc = system(os(...) ...) then
            {System.show Exc}
            {System.printError "Can't find input file: "#"\n"}
            {Application.exit 2}
         end
      end
   catch X then
      case X of error(ap(usage M) ...) then
         AppURL = {Property.get 'application.url'} in
         {System.printError
          'Command line option error: '#M#'\n\n'#
          'Usage: '#AppURL#' <GIF-Files> [options] \n'#
          '  -o , --output   <OutputFile>      (default: Gifs.ozf)\n'#
          '  -z , --compress <CompressLevel>   (0-9, default: 0)\n'}
	 {Application.exit 2}

      elseof E then
	 {Browser.browse E}
%	 raise E end
      end
   end
end

/*

declare
[F]={Module.link ['Gifs.ozf']}
T={New Tk.toplevel tkInit()}
L={New Tk.label tkInit(parent:T image:F.image)}
{Tk.send pack(L)}

*/
