%%% Author:
%%%   Andreas Franke <afranke@ags.uni-sb.de>
%%%
%%% Copyright:
%%%   Andreas Franke, 2000
%%%
%%% Last change:
%%%   $Date: 2001/07/30 21:03:11 $ by $Author: afranke $
%%%   $Revision: 1.1 $
%%%
%%% This file is part of MathWeb
%%%    http://www.mathweb.org
%%%
%%% See the file "LICENSE" or
%%%    http://www.mathweb.org/mathweb/LICENSE.html
%%% for information on usage and redistribution
%%% of this file, and for a DISCLAIMER OF ALL
%%% WARRANTIES.
%%%

functor
   
import
   Application(getCmdArgs exit)
   System(showError)
   Pickle(loadWithHeader saveWithHeader)
   
define
   AppName = 'ozzip'

   %%
   %% auxiliuary procedures
   %%
   proc {ZipFile InFile OutFile Level}
      Header#Val = {Pickle.loadWithHeader InFile} in
      {Pickle.saveWithHeader Val OutFile Header Level}
   end

   %%
   %% argument processing
   %%
   proc {Usage}
      {System.showError 
       'Usage: ozzip <InputFile> [-o <OutputFile>] [-z <Compression>]'}
      {System.showError 
       'or:    ozzip <InputFile1> <InputFile2> [...] [-z <Compression>]'}
      {Application.exit 1}
   end
   
   Spec = record('out'(single char:&o type:string optional:true)
                 'compress'(single char:&z type:int(min:0 max:9) default:9)
                 'level'(single alias:'compress'))
   Args
   try
      Args={Application.getCmdArgs Spec}
   catch error(ap(usage S)) then
      {System.showError AppName#": "#S}
      {Application.exit 1}
   end

   Level = Args.compress

   %%
   %% main
   %%
   case Args.1
   of nil then {Usage}
   [] [InFile] then
      OutFile = {CondSelect Args out InFile} in
      {ZipFile InFile OutFile Level}
   [] InFiles then
      if {HasFeature Args out}
      then {Usage}
      else {ForAll InFiles proc {$ F} {ZipFile F F Level} end}
      end
   end
   
   {Application.exit 0}
end

